"""

Utilities for NCMAT creation, most notably the NCMATComposer helper class which
provides a pythonic and flexible object oriented interface for creating NCMAT
data representing a given material, and for verifying, writing, registering, or
loading said data.

"""

################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

class NCMATComposer:
    """Helper class for composing an NCrystal material.

    One can either start with an empty instance, or bootstrap the process by
    starting the material from an existing NCMAT file, an NCrystal cfg-string,
    an NCrystal.Info object, a CIF file, or from a variety of files (anything
    that can be loaded with ASE and contains a suitable crystal
    structure). Depending on the specifics of the problem, one will then use
    various methods on the object to add or modify material information as
    desired, before ultimately the create_ncmat() method can be used to create
    the resulting NCMAT data. As a convenience the .write() and .register_as()
    methods can be used to write this data as a physical or virtual file
    respectively, or the .load() method can be used to directly load the
    data. Finally, the .plot_xsect() and .inspect() methods can also be used to
    quickly load and investigate the material. It might also be useful to print
    the output of .create_ncmat(), to understand what material has been created.

    For a more in-depth discussion and usage examples, please refer to the
    example Jupyter-Lab notebooks in the
    https://github.com/mctools/ncrystal-notebooks repository.

    Note that the "labels" used in the various methods, to identify different
    atoms in the materials are custom user labels. They are simply a way to
    provide a handle associated with a particular atom role or species in the
    material. For many simple materials, all labels will simply have the name of
    the element or isotope ("Al", "Li6", "D", ...), and each type of element and
    isotope will only appear once in the material. That is certainly OK, for
    such simple materials, and in these cases there is no need to concern
    one-self with the atomic composition of each label, since it can be inferred
    from the label itself. However, sometimes different atoms of the same type
    will appear in more than one role in the material (i.e. there might be one
    group of loosely bound aluminium atoms, while a different group of aluminium
    atoms are more tightly bound). Or, it might be that a given label is
    actually associated with an actual mix of elements of isotopes (i.e. a label
    might be associated with a particular isotopic enrichment, or a
    multi-element mixture). In these cases, each label should have it's precise
    composition defined with a call to the .set_composition(..) method.

    In other words: Unless the label happens to coincide with the name of an
    element or isotope ("Al", "Li6", "D", ...), its atomic composition MUST be
    defined by a call to the .set_composition(..) method. And if a given element
    has more than one role in the material, it must of course also have more
    than one label - one for each role (e.g. "tightAl" and "looseAl").

    To ensure high data quality and reliability, all crystalline materials will
    by default have their crystal structure verified for consistency with the
    indicated spacegroup, before it is converted to NCMAT data. For that reason,
    the third-party module "spglib" must be installed in order for this
    verification to take place (note that the .refine_crystal_structure() method
    can also be invoked before the material is used, to detect the spacegroup in
    if it was not provided). Fortunately, "spglib" is available both on PyPI and
    in the conda-forge conda channel.

    """

    def __init__(self, data = None, fmt = None, quiet = False, plotlabel = None ):
        """Initialise. Either an empty instance, or if data (and possibly fmt)
        is provided, from a variety of data. Typically, the fmt parameter can be
        left out, but occasionally it might be needed to specify it
        explicitly. Supported fmt's are "cif", "cfgstr", "ncmat", (for NCMAT
        data), "info" (for NCrystal.Info or LoadedMaterial objects), ase (for
        ASE structures). Finally, the format "via_ase" is special, since it will
        always for the input to be processed with ASE, even if another more
        obvious format exists (i.e. CIF files will be processed via ASE first,
        not passed along directly to the NCrystal.CIFLoader). As an alternative
        to the "fmt" parameter, one can also invoke the various .from_xxx(..)
        static methods. This has the potential advantage of allowing the
        specification of parameters dedicated only to that input data type.

        If quiet=True, no informative messages will be emitted, and the plotlabel
        will be passed along to .set_plotlabel().
        """
        from ._ncmatimpl import NCMATComposerImpl as Impl
        self.__impl = ( data if isinstance( data, Impl )
                        else ( data.__impl if isinstance( data, NCMATComposer )
                               else Impl( data = data, fmt = fmt, quiet = quiet ) ) )
        if plotlabel:
            self.set_plotlabel(plotlabel)

    @staticmethod
    def from_cif( cifsrc, quiet=False, mp_apikey = None,
                  uiso_temperature = None,
                  override_spacegroup = None, **kwargs ):
        """Initialise from CIF data (anything suitable for the
        NCrystal.cifutils.CIFSource class). Internally, the loading is carried
        out by instantiating a CIFLoader instance (from the NCrystal.cifutils
        module), and calling it's .create_ncmat_composer() method. Any provided
        arguments will be passsed along to the call to .create_ncmat_composer(),
        except for the override_spacegroup argument which is used when the
        CIFLoader is instantiated.

        For advanced work, it might be better to use the CIFLoader class
        directly.
        """
        from ._ncmatimpl import NCMATComposerImpl as Impl
        return NCMATComposer( Impl.from_cif( cifsrc=cifsrc, quiet=quiet,
                                             uiso_temperature = uiso_temperature,
                                             override_spacegroup = override_spacegroup,
                                             mp_apikey = mp_apikey, **kwargs ) )

    @staticmethod
    def from_cfgstr( cfgstr ):
        """Loads an Info object from the given cfg-string, and pass it on to
        .from_info(..)."""
        from ._ncmatimpl import NCMATComposerImpl as Impl
        return NCMATComposer( Impl.from_cfgstr( cfgstr = cfgstr ) )

    @staticmethod
    def from_ase( ase_obj, ase_format = None, *, quiet = False, **kwargs ):
        """Initialise from an ASE object. This is actually done internally by
        first converting the ASE object to CIF data, and then subsequently
        passing it on to the .from_cif(..) method (any **kwargs will be
        passed along to the .from_cif call). If the provided ase_obj is not
        actually already an ASE object, an attempt is done to load it into one
        via the ase.io.read module (mimicking the fmt="via_ase" behaviour of the
        __init__ method). In that case, any value in the ase_format parameter
        will be passed along to ase.io.read's format parameter.
        """
        from ._ncmatimpl import NCMATComposerImpl as Impl
        return NCMATComposer( Impl.from_ase( ase_obj = ase_obj, quiet = quiet,
                                             ase_format = ase_format, **kwargs ) )

    @staticmethod
    def from_info( info_obj ):
        """Initialise from provided Info (or LoadedMaterial) object.

        In case the material is a gas, the temperature of the resulting
        NCMATComposer will for safety be locked (due to the strong temperature
        dependency of densities for gasses).
        """
        from ._ncmatimpl import NCMATComposerImpl as Impl
        return NCMATComposer( Impl.from_info( info_obj = info_obj ) )

    @staticmethod
    def from_ncmat( data ):
        """Loads an Info object from the given NCMAT data, and passes it on to
        .from_info(..). If the NCMAT data has no newlines and doesn't start with
        'NCMAT', it will actually be passed directly on to .from_cfgstr(..)
        instead.
        """
        from ._ncmatimpl import NCMATComposerImpl as Impl
        return NCMATComposer( Impl.from_ncmat( data ) )

    def __call__( self, cfg_params = None, **kwargs ):
        """Convenience short-cut for the .create_ncmat() method"""
        return self.create_ncmat( cfg_params = cfg_params, **kwargs )

    def update_atomdb( self, element_or_isotope, data = None, *,  mass = None, coh_scat_len=None, incoh_xs=None, abs_xs=None ):
        """Specify atom data for particular element or isotope. Either specify
        the data parameter as a string (in @ATOMDB format like "28.97649466525u
        4.7fm 0.001b 0.101b") or AtomData object, or put all four physics values
        (mass, coh_scat_len, incoh_xs, and abs_xs) directly as floating point
        numbers. More specifically these numbers are mass value (amu), coherent
        scattering length (fm), incoherent cross section (barn), and absorption
        cross section @v_n=2200m/s (barn).
        """
        return self.__impl.update_atomdb( element_or_isotope,data=data,mass=mass,
                                          coh_scat_len=coh_scat_len,incoh_xs=incoh_xs,abs_xs=abs_xs )

    def find_label( self, element, allow_multi=False ):
        """Attempt to locate label which has a composition containing provided
        element (which can be a string name or Z value like Al, 13, B10,
        ...). If exactly one such label exists, it will be returned, otherwise
        None is returned. If allow_multi==True, the result(s) will instead be
        returned as a list, and all matching labels will be returned.
        """
        return self.__impl.find_label( element = element, allow_multi = allow_multi )

    def set_composition( self, label, *composition ):
        """Use like .set_composition('H','D') or .set_composition('H is D'). Mixtures
        can also be defined, for instance like .set_composition('H','0.2 H1 0.8
        H2'). For a more in-depth discussion of the syntax, consult the @ATOMDB
        section on https://github.com/mctools/ncrystal/wiki/NCMAT-format

        Set atomic composition associated with a given label. Usage examples:

        .set_composition('mylbl','Al') #mylbl now has a composition of only (natural) Al atoms.
        .set_composition('Al','Al') #Al atoms (notice this call is not needed
                                    #since the label is already an element name).
        .set_composition('Al','Cr') #Well if you want to use confusing labels you can
        .set_composition('mylbl','Li6') #specific isotope
        .set_composition('mylbl','D') #deuterium is supported by either 'D' or 'H2'

        More complicated mixtures are supported as well:

        .set_composition('mylbl',0.95,'B10',0.05,'B11')#enriched boron
        .set_composition('mylbl',[0.95,'B10',0.05,'B11'])#same but defined via sequence
        .set_composition('mylbl',[(0.95,'B10'),(0.05,'B11')])#same but defined via sequence of pairs
        .set_composition('mylbl','0.95 B10 0.05 B11')#same but via a string in @ATOMDB-like form.
        .set_composition('mylbl is 0.95 B10 0.05 B11')#same but even more @ATOMDB like.
        .set_composition('Al is 0.99 Al 0.005 Cr 0.005 B10')#mix and match elements and isotopes
        """
        return self.__impl.set_composition( label, *composition )

    def remap_atom( self,  element_or_isotope, *composition ):
        """Calling this method updates all usage of the denoted
        element_or_isotope marker ('C', 'Al', 'D', 'B10', ...) and replaces it
        with the particular composition. For instance, .remap_atom('H','D') can
        be used to fully deuterate a material. For a description of the allowed
        compositon syntax, refer to the .set_composition method.

        The remapping will only affect labels and compositions that have already
        been added to the NCMATComposer object.
        """
        return self.__impl.remap_atom( element_or_isotope, *composition )

    def clear_comments( self ):
        """Clear comments added with the add_comments method."""
        return self.__impl.clear_comments()

    def add_comments( self, comments, add_empty_line_divider=False ):
        """Add comments (string or list of strings) to the top of NCMAT data."""
        return self.__impl.add_comments( comments = comments,
                                         add_empty_line_divider = add_empty_line_divider )

    def clone(self):
        """Return a new independent NCMATComposer object which is a clone of the
        current one."""
        return NCMATComposer( self.__impl.clone() )

    def get_cache_key( self ):
        """
        Returns tuple of integers which will be unchanged if generated NCMAT
        content is unchanged. Note that this is partially based on memory
        addresses of python objects, so the keys are NOT reproducible outside
        the current process.
        """
        return self.__impl.get_cache_key()

    def to_dict(self):
        return self.__impl.to_dict()

    def set_cellsg( self, *, a,b,c, alpha,beta,gamma, spacegroup=None ):
        """Assumes alpha,beta,gamma in degrees, a,b,c in Aangstrom, spacegroup
        an int in range 1..230."""
        return self.__impl.set_cellsg( a=a,b=b,c=c,
                                       alpha=alpha,beta=beta,gamma=gamma,
                                       spacegroup=spacegroup )

    def set_cellsg_cubic( self, a, *, spacegroup=None ):
        """Like .set_cellsg but will set b=c=a and alpha=beta=gamma=90."""
        return self.__impl.set_cellsg_cubic( a = a, spacegroup = spacegroup )

    def set_atompos( self, atompos ):
        """The atompos parameter must be list of entries
        (label,x,y,z) or (label,x,y,z,site_occupancy).

        If provided, the site_occupancy value (a number in (0,1] must be the
        same for all identical labels.

        The site_occupancy parameter is for now highly experimental and will
        result in generated NCMAT files with fake sterile atoms inserted for
        technical reasons (resulting in a higher density but hopefully correct
        physics as long as the user does not subsequently try to override the
        density directly via the cfg-string parameters). This is perfectly fine
        for usage in a MC simulation context where NCrystal provides the entire
        physics of the material (e.g. McStas), but in a context where a base
        material must be created using the composition from NCrystal
        (e.g. Geant4 or OpenMC), it most likely will result in an error.
        """
        return self.__impl.set_atompos( atompos = atompos )

    def refine_crystal_structure( self, symprec = 0.01, quiet = False ):
        """Attempt to refine the crystal structure (does nothing if not a
        crystalline material). This ignores the spacegroup number (if any), and
        uses spglib to refine and standardise the crystal structure based on the
        provided unit cell parameters and atomic positions. The unit cell
        parameters, spacegroup number, and atomic positions, are all updated
        accordingly.
        """
        return self.__impl.refine_crystal_structure( symprec = symprec,
                                                     quiet = quiet )

    def verify_crystal_structure( self, symprec = 0.01, quiet = False ):
        """Attempt to verify the crystal structure (does nothing if not a
        crystalline material). This requires a spacegroup number to be
        available, and uses spglib to verify that the crystal structure (unit
        cell and atom positions) is consistent with that spacegroup
        number. Raises an exception in case the structure could not be verified,
        otherwise does nothing.
        """
        return self.__impl.verify_crystal_structure( symprec = symprec,
                                                     quiet = quiet )

    def set_density( self, value, unit = 'g/cm3' ):
        """
        Specify material density for non-crystalline materials. If provided,
        the unit must be a string which is one of "g/cm3" (default) , "kg/m3",
        or "atoms/Aa3". Note that any material can have its density modified
        with the cfg-string parameter "density", while the present method is
        instead used to provide the basic density embedded in NCMAT format for
        non-crystalline materials.
        """
        return self.__impl.set_density( value = value, unit = unit )

    def set_fraction( self, label, value ):
        """
        Specify fraction of component associated with a given label. This is
        not needed for crystalline materials, where the fractions can be
        inferred from the unit cell contents. Note that the various
        set_dyninfo_xxx methods have an optional fraction parameter, which can
        also be used to set the fraction directly.

        """
        return self.__impl.set_fraction( label = label, value = value )

    def allow_fallback_dyninfo( self, debye_temp = 300.0 ):
        """For crystalline materials only, allow a fall-back modelling of any
        components with absent dynamic information via a VDOS-Debye model using
        the specified Debye temperature.
        """
        return self.__impl.allow_fallback_dyninfo( debye_temp = debye_temp )

    def set_dyninfo_vdos( self, label, vdos_egrid, vdos, *, fraction = None, comment = None ):
        """Set dynamics of component to be modelled by a 1D phonon density of
        state (DOS) curve (VDOS="Vibrational DOS"). This is the preferred way to
        provide dynamics for components of solid materials, as it not only
        gives enough information to estimate the Debye-Waller factors needed for
        elastic scattering in all such materials, but also allows realistic
        inelastic scattering via expansion into full 2D scattering kernels.
        This model is not appropriate for gaseous or liquid materials.
        """
        return self.__impl.set_dyninfo_vdos( label = label,
                                             vdos_egrid = vdos_egrid,
                                             vdos = vdos,
                                             comment = comment,
                                             fraction = fraction )

    def set_dyninfo_vdosdebye( self, label, debye_temp, *, fraction = None, comment = None ):
        """Set dynamics of component to be modelled by a Debye temperature. This
        allows temperature-dependent atomic displacements and Debye-Waller
        factors to be estimated, and inelastic scattering will be modelled based
        on an idealised VDOS curve (a parabola up to a cutoff frequency
        corresponding to E=k*T_debye). This model is only appropriate for solid
        materials.
        """
        return self.__impl.set_dyninfo_vdosdebye( label = label,
                                                  debye_temp = debye_temp,
                                                  comment = comment,
                                                  fraction = fraction )

    def set_dyninfo_debyetemp( self, label, debye_temp, *, fraction = None, comment = None ):
        """Alias for set_dyninfo_vdosdebye."""
        return self.set_dyninfo_vdosdebye( label=label, debye_temp=debye_temp,
                                           fraction=fraction, comment=comment )

    def set_dyninfo_msd( self, label, msd, *, temperature, fraction = None, comment = None,  ):
        """Calculate and set Debye temperature based on
        mean-squared-displacement (msd) value (in Aa^2). This also needs the
        temperature value for which the msd value is associated (in
        kelvin). Note that this temperature value is NOT necessarily the
        temperature at which the material will later be used. See
        .set_dyninfo_vdosdebye(..) for further information about the resulting
        modelling.
        """
        return self.__impl.set_dyninfo_msd( label = label,
                                            msd = msd,
                                            temperature = temperature,
                                            comment = comment,
                                            fraction = fraction )

    def set_dyninfo_uiso( self, label, uiso, temperature, *, fraction = None, comment = None ):
        """Alias for set_dyninfo_msd. The uiso value is just another name for
        "msd" and both have a unit of Aa^2."""
        return self.set_dyninfo_msd( label=label, msd=uiso, temperature=temperature,
                                     fraction=fraction, comment=comment )

    def set_dyninfo_scatknl( self, label, *, alphagrid, betagrid, temperature,
                             sab = None, sab_scaled = None, egrid = None,
                             fraction = None, comment = None ):
        """Set dynamics of component to be modelled by the provided 2D
        S(alpha,beta) ("sab") scattering kernel. Such kernels are
        temperature-dependent, so one must also specify the temperature for
        which the kernel is valid, which of necessity will lock the temperature
        of the entire material at that temperature. As NCrystal currently is not
        able to estimate atomic displacements and Debye-Waller factors from such
        kernels, they can not be used with crystalline materials at all, and are
        in principle not suitable for amorphous solids either, as even
        incoherent-elastic contributions will be absent. Thus, this modelling is
        mainly intended for liquids or gasses.
        """
        return self.__impl.set_dyninfo_scatknl( label = label,
                                                alphagrid = alphagrid,
                                                betagrid = betagrid,
                                                temperature =  temperature,
                                                sab = sab,
                                                sab_scaled = sab_scaled,
                                                egrid = egrid,
                                                comment = comment,
                                                fraction = fraction )

    def set_dyninfo_freegas( self, label, *, fraction = None, comment = None ):
        """Set dynamics of component to be modelled by a free gas description (a gas
        of non-interacting atoms). This model is not suitable for solid
        materials, except as a crude approximation. It is specifically
        disallowed for crystalline materials, since it does not provide atomic
        displacements (Debye-Waller factors). See .set_dyninfo_debyetemp(..),
        .set_dyninfo_msd(..) and .allow_fallback_dyninfo() for easy
        (low-realism) alternatives for crystalline materials.
        """
        return self.__impl.set_dyninfo_freegas( label = label,
                                                comment = comment,
                                                fraction = fraction )

    def set_dyninfo_sterile( self, label, *, fraction = None, comment = None ):
        """Set dynamics of component to be modelled as being without any scattering
        cross section. This model is obviously unrealistic and exists for
        debugging purposes only. It can not be used for crystalline materials
        (for these one could use .update_atomdb instead to introduce atoms with
        no scattering cross section into the crystal).
        """
        return self.__impl.set_dyninfo_sterile( label = label,
                                                comment = comment,
                                                fraction = fraction )

    def set_dyninfo_from_object( self, label, source_dyninfo, comment = None, fraction=None ):
        """Set dyninfo of given label based on existing NCrystal.DynamicInfo
        object.

        Note that in principle the flow NCMAT-data -> Info -> NCMATComposer ->
        NCMAT-data is not guaranteed to be 100% lossless, but in practice the
        resulting physics is unlikely to be unaffected by this. For a guaranteed
        loss-less operation, one must manually edit the NCMAT data and copy the
        exact lines of the relevant @DYNINFO section into the new NCMAT data
        (this is beyond the scope of the NCMATComposer).
        """
        return self.__impl.set_dyninfo_from_object( label = label,
                                                    source_dyninfo = source_dyninfo,
                                                    comment = comment,
                                                    fraction = fraction )

    def transfer_dyninfo_objects( self, source, mapping = None, allow_none = False ):
        """Set dyninfo from source. Source can be a list of
        NCrystal.DynamicInfo, a single NCrystal.DynamicInfo info object, or any
        sort of material source like an Info object, a cfg-str, NCMAT data, etc.

        If a mapping dict is provided, it must be a mapping between labels used
        on the NCMATComposer object and display labels in the source DynamicInfo
        objects (example: {"H":"H","mylbl":"Al","O16":"O"}). If a mapping is NOT
        provided, an automatic mapping will be attempted based on element Z
        values between source entries and labels already registered in the
        composer (in this case it must be possible to determine the composition
        of all labels). Any excess entries in the source will be ignored.

        An NCBadInput error will be raised if no dynamic information is actually
        transferred, unless allow_none=True is provided.

        """
        return self.__impl.transfer_dyninfo_objects( source = source,
                                                     mapping = mapping,
                                                     allow_none = allow_none )

    def set_state_of_matter( self, state_of_matter ):
        """Explicitly specify the state of matter, which must be a string with
        value of "solid", "liquid", or "gas". Call it with an empty string or
        None to clear the information.

        Note that NCrystal will automatically classify materials as "solid" if
        they are crystalline materials and/or have dynamic information of the
        types vdos, vdosdebye, or msd. Trying to change the state of matter
        value for such materials will result in an error at a later stage.
        """
        return self.__impl.set_state_of_matter( state_of_matter )


    def add_secondary_phase(self, fraction, cfgstr, normalise = True ):
        """Add a secondary phase to the material, by specifying the phase volume
        fraction and a cfg-strings defining the content of the secondary
        phase. For instance a material with air bubbles inside, might be
        emulated by calling .add_secondary_phase(0.05,"gasmix::air"), which
        would add a secondary phase to the material, occupying 5% of the volume
        and containing air. For details, please see the discussion under the
        heading "The @OTHERPHASES section" in the document at
        https://github.com/mctools/ncrystal/wiki/NCMAT-format.

        Unless normalise=False, the cfg-string will be normalised with a call to
        normaliseCfg(..) from the NCrystal.cfgstr module.
        """
        return self.__impl.add_secondary_phase( fraction = fraction,
                                                cfgstr = cfgstr,
                                                normalise = normalise )

    def add_hard_sphere_sans_model( self, sphere_radius ):
        """As a technology preview of future more complete SANS support, one can
        for now use this method to add hard-sphere SANS scattering between the
        primary phase, and the first phase added with .add_secondary_phase(..)
        Note that as this results in a @CUSTOM_HARDSPHERESANS section in the
        resulting NCMAT data (because it is a tech-preview), warnings will be
        emitted when the data is subsequently loaded by NCrystal.

        The implemented hard-sphere model only has a single free parameter, the
        sphere_radius in angstrom.
        """
        return self.__impl.add_hard_sphere_sans_model(sphere_radius=sphere_radius)

    def lock_temperature( self, value ):
        """
        Lock the temperature of the material to the given value. This not
        only changes the default temperature of the material, but also "locks"
        it in the sense that any attempts at using the cfg-level variable "temp"
        to modify the temperature will result in an error. See also the
        .set_default_temperature() method. Call with value=None to clear any
        previous effects of .lock_temperature() or .set_default_temperature().
        """
        return self.__impl.lock_temperature( value )

    def set_default_temperature( self, value ):
        """
        Modify the default temperature of the material from the usual
        293.15K. The temperature value can still be modified subsequently using
        the cfg parameter "temp" as usual. To prevent this, use the
        .lock_temperature() method instead. Call with value=None to clear any
        previous effects of .lock_temperature() or .set_default_temperature().
        """
        return self.__impl.set_default_temperature( value )

    def get_temperature_setting( self ):
        """Returns the temperature setting resulting from calls to
        .lock_temperature() or .set_default_temperature(). Returns a tuple of
        two values: the temperature value in kelvin, and a boolean indicating
        whether the value has been locked or is merely a changed
        default. Returns (None,None) in the absence of such a setting.
        """
        return self.__impl.get_temperature_setting()

    @property
    def state_of_matter( self ):
        """
        Access the state of matter which will be one of "solid", "liquid",
        or "gas". However, this is only available if explicitly set, otherwise
        None is returned.
        """
        return self.__impl.get_state_of_matter()

    def get_labels( self ):
        """List of all labels registered so far."""
        return self.__impl.get_labels()

    def write( self, path, cfg_params = None ):
        """Produce NCMAT data and write it to the provided file path. If any
        cfg_params are provided, they will be embedded in the NCMAT data using
        the NCRYSTALMATCFG[..] syntax.
        """
        return self.__impl.write( path = path,
                                  cfg_params = cfg_params )

    def register_as( self, virtual_filename, cfg_params = None ):
        """Produce NCMAT data and register it in memory with the provided
        virtual file-name, using the registerInMemoryFileData(..) function from
        the NCrystal.datasrc module. If any cfg_params are provided, they will
        be embedded in the NCMAT data using the NCRYSTALMATCFG[..] syntax.
        """
        return self.__impl.register_as( virtual_filename = virtual_filename,
                                        cfg_params = cfg_params )

    def load(self, cfg_params = None, *, force = False ):
        """Will create NCMAT and load it with the directLoad(..) function of the
        NCrystal.core module. The cfg_params parameter can be used to apply cfg
        parameters (e.g. cfg_params="temp=200K;dcutoff=0.1")., while force=True
        will prevent the NCMATComposer from simply returning a previously loaded
        material (usually one should just leave the default force=False value
        untouched).
        """
        return self.__impl.load( cfg_params = cfg_params, force = force )

    def plot_xsect( self, cfg_params = None, **kwargs_plot_xsect ):
        """
        Quick plot (with matplotlib) showing the cross sections produced by
        current material, possibly after appending certain cfg_params. This is
        using the plot_xsect function from the NCrystal.plot module, so refer to
        that function for available arguments.
        """
        return self.__impl.plot_xsect( self, cfg_params, kwargs_plot_xsect )

    def inspect( self, cfg_params = None, **kwargs_plot_xsect ):
        """Similar to the plot_xsect(..) function, but will also print out
        information (e.g. .dump()) about the Info, Scatter, and Absorption
        objects.
        """
        return self.__impl.inspect( self, cfg_params, kwargs_plot_xsect )

    def get_chemical_composition( self, as_str = False ):
        """Chemical composition as list of [(elemiso,count),...] where elemiso
        is a marker like "Al", "D", "B10", .... For crystals, the counts will
        represent the number of elements/isotopes per unit cell. Returns None in
        case of incomplete information.

        If as_str = True, the result will instead be returned encoded in a
        string (e.g. "H2O").
        """
        return self.__impl.get_chemical_composition( as_str=as_str )

    def create_ncmat( self, cfg_params = None, *,
                      meta_data = False,
                      verify_crystal_structure = True ):
        """Creates and returns NCMAT data based on the material settings
        provided so far. Any cfg_params provided here will be embedded into the
        NCMAT data itself, using the NCRYSTALMATCFG[..] syntax.

        If meta_data=True, the return value will be a tuple whose first element
        is the created NCMAT data, and whose second element is a dictionary with
        a high-level meta-data concerning the material (like chemical
        composition, spacegroup, etc.).

        Unless verify_crystal_structure=False (not recommended!) any crystalline
        material will have its structure verified by calling
        .verify_crystal_structure(), possibly raising an exception in case of
        issues found. Depending on the scenario, it might be possible to prevent
        this by calling .refine_crystal_structure(), which will hopefully fix it
        up before verification (but be vigilant and double-check that the
        refinement did not simply hide some fundamental flaw of the input data).
        """
        return self.__impl.create_ncmat( cfg_params = cfg_params,
                                         meta_data = meta_data,
                                         verify_crystal_structure = verify_crystal_structure )


    def set_plotlabel( self, lbl ):
        """Sets the plotlabel (cf. .plotlabel)."""
        return self.__impl.set_plotlabel(lbl)

    @property
    def plotlabel( self ):
        """An optional label for the material, which can be used for plotting
        purposes in legends, etc.
        """
        return self.__impl.plotlabel

    def as_spglib_cell( self ):
        """For a crystalline material, return the unit cell definition in a
        format suitable for usage in spglib calls."""
        return self.__impl.as_spglib_cell()

    def _unofficial_vdos2sab_ignore( self, *, order_low, order_high = None, mode = None ):
        """This expert-only method can be used to exclude certain phonon orders
        from the scattering kernel during vdos2sab expansion. It is primarily
        intended as a debugging aid for visualisation purposes, or for
        developers intending to develop coherent single-phonon physics.

        Be aware that, for now, trying to use the Info.DI_VDOS[Debye].loadKernel
        or .plot_knl(), will NOT reflect this hack!
        """
        return self.__impl._unofficial_vdos2sab_ignore( order_low = order_low,
                                                        order_high = order_high,
                                                        mode = mode )

    def _get_impl_obj( self ):
        return self.__impl

def formatVectorForNCMAT(name,values,indent='  '):
    """Utility function for help in python scripts composing .ncmat files,
       transforming an array of of values into a properly formatted text string,
       with word-wrapping, usage of <val>r<n> syntax, etc.
    """
    from ._ncmatimpl import formatVectorForNCMAT as _
    return _( name = name,
              values = values,
              indent = indent )

def _rawParseNCMAT(text_data_name,*,asJSONStr=False):
    """Parses NCMAT content and returns as Python data structure (a dictionary). The
       format of this data structure should be mostly self-evident by
       inspection, and is not guaranteed to stay the same across NCrystal
       versions. If asJSONStr=true, the data structure will be returned as a
       JSON-encoded string, instead of a Python dictionary.

       WARNING: This function is considered experimental and is currently NOT
       feature complete. It only returns data from a few select NCMAT sections."""
    from ._chooks import _get_raw_cfcts
    _js = _get_raw_cfcts()['nc_ncmat2json'](text_data_name if not hasattr(text_data_name,'rawData') else text_data_name.rawData )
    if asJSONStr:
        return _js
    import json
    return json.loads(_js)




################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

"""

Utilities related to plugins and dynamic factories.

"""

def hasFactory(name):
    """Check if a factory of a given name exists"""
    from ._chooks import _get_raw_cfcts,_str2cstr
    return bool(_get_raw_cfcts()['ncrystal_has_factory'](_str2cstr(name)))

def browsePlugins(dump=False):

    """Return list of plugins [(pluginname,filename,plugintype),...].

    If the dump flag is set to True, the list will not be returned. Instead it
    will be printed to stdout.
    """
    from ._chooks import _get_raw_cfcts
    l=_get_raw_cfcts()['ncrystal_get_pluginlist']()
    if not dump:
        return l
    from ._common import print
    print('NCrystal has %i plugins loaded.'%len(l))
    for i in range(len(l)):
        pluginname, filename, plugintype = l[i]
        print('==> %s (%s%s)'%(pluginname,plugintype,
                             ' from %s'%filename if filename else ''))




################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

"""Various utility functions and classes that do not fit elsewhere.

Of particular note, this module provides several type-erasure classes which can
facilitate the coding of rich and convenient interfaces able to accept many
input formats. For instance, the concept of "a material" in NCrystal could be
either an .ncmat file, a cfg-str, an in-memory string or an NCrystal.TextData
instance with NCMAT data, or even pre-loaded objects like Info, Scatter,
Absorption, or LoadedMaterial objects. Writing a function which works on "a
material" could therefore be quite cumbersome, but by passing the input through
the MaterialSource type erasure class from this module, all of these inputs can
be handled using a single interface.

Similar type-erasure classes exists for other types of data as well (e.g. VDOS
curves, generic text data, ...).

"""

#For client code checking all types of components:
standard_comp_types = ('coh_elas','incoh_elas','inelas','sans')

def detect_scattering_components( material ):
    """Detect which scattering components are present in the material.

    More precisely, the component names listed in the standard_comp_types
    variable will be checked, and a list of those present will be returned. Note
    they are suitable for subsequent usage in the "comp" cfg-string
    parameter.

    Note: This detection only works for material sources which are not
    preloaded. Trying to use this function with a preloaded material sources
    (e.g. an NCrystal.Scatter object) will result in an error.
    """
    from ._miscimpl import detect_scatcomps as _
    return _( standard_comp_types, MaterialSource( material ) )

class MaterialSource:
    """
    Generic type-erasure class for NCrystal materials in a general sence. It can
    be used to wrap cfg-strings, NCMAT data, TextData, LoadedMaterial objects,
    or even individual Info / Scatter / Absorption objects, etc.

    The most important feature of a MaterialSource object is that it has a
    .load(..) method which can be used to load the material into a
    LoadedMaterial object. Some MaterialSource objects will "preloaded", in the
    sense that they wrap already loaded objects (Info, Scatter, Absorption, or
    LoadedMaterial objects). MaterialSource objects that are NOT "preloaded"
    will have to do some work when the .load(..) method is invoked, but can on
    the other hand also offer a bit more flexibility (in particular the
    possiblity for loading with extra cfg-string parameters "appended").
    """

    def __init__( self, data, *, fmt = None, cfg_params = None ):
        """Wrap data object. If the data object is not a preloaded physics
        object, it is also allowed to specify cfg_params like
        cfg_params="temp=200K;vdoslux=2". Finally, the fmt string will be
        detected automatically, unless the fmt parameter is provided.
        """
        from . import _miscimpl
        self.__d = _miscimpl.matsrc_init( MaterialSource, lambda matsrc : matsrc.__d,
                                          data,cfg_params=cfg_params,fmt=fmt)

    @property
    def description( self ):
        """Returns a string with a MaterialSource description."""
        return self.__d['description']

    @property
    def plotlabel( self ):
        """Returns a string with either a specific plotlabel or otherwise just the usual description."""
        return self.__d.get('plotlabel') or self.__d['description']

    def __str__(self):
        return 'MaterialSource(%s)'%self.description

    def __repr__(self):
        return str(self)

    @property
    def is_preloaded( self ):
        """
        Check if material is already loaded into NCrystal physics
        objects. If so, the .load() function will be very cheap to call, but can not
        accept any extra_cfg_params.
        """
        return bool( self.__d.get('preloaded') )

    def load( self, extra_cfg_params = None, doInfo = True, doScatter = True, doAbsorption = True ):
        """
        Loads material (if not preloaded) and in any case returns a
        LoadedMaterial object. If material is preloaded, extra_cfg_params can
        NOT be provided, and the doInfo/doScatter/doAbsorption flags will have
        no effect.
        """
        ecp = (extra_cfg_params or '').strip()
        preloaded = self.__d.get('preloaded')
        if ecp and preloaded:
            from .exceptions import NCBadInput
            raise NCBadInput( 'Can not accept extra_cfg_params on a'
                              ' MaterialSource which is preloaded.' )
        if preloaded:
            return preloaded
        return self.__d['loadfct']( extra_cfg = extra_cfg_params,
                                    doInfo = doInfo,
                                    doScatter = doScatter,
                                    doAbsorption = doAbsorption )

class AnyTextData:
    """
    Immutable type-erasure class for text data (possibly named).
    Be aware, that this class will always keep the entire content in memory.
    """

    def __init__( self, data, is_path = None, name = None ):
        """
        Initialise from data, which can be either in-memory text data (str
        or NCrystal.TextData) or a file path (str or pathlib.Path) from which
        data will be loaded. Data of type bytes will simply be attempted decoded
        as a string.

        A string containing no new-line characters is assumed to be a file-path,
        otherwise it will be interpreted as in-memory data. Set is_path to True
        or False to override this logic.

        The "name" parameter can be used to specify or override the "name" of
        the file source.
        """
        if isinstance(data,AnyTextData):
            _ = ( name or data.__name ), data.__content, data.__keepalive
        else:
            from . import _miscimpl
            _ = _miscimpl.anytextdata_init( data, is_path=is_path, name=name )
        self.__name, self.__content, self.__keepalive = _

    @property
    def name( self ):
        """Access a name of the text data (could for instance be the file name). None if not available."""
        return self.__name

    @property
    def content( self ):
        """Access the contents (always type str)."""
        return self.__content

class AnyVDOS:
    """
    Immutable type-erasure class for phonon DOS (VDOS) curves.
    """

    def __init__( self, data, fmt = None, label = None ):
        """Initialise from data in various formats. Currently, input data can
        either be raw curves in the form of two arrays (egrid,density), DI_VDOS
        objects, or DI_VDOSDebye objects. Raw curves with egrid[0]>0 will be
        extended with a parabola towards (0,0). Furthermore egrid values
        consisting of just two points will be assumed to be a linearly spaced
        grid with those endpoints.

        The label parameter can be used to assign a label to the curves, which
        might for instance be used in utilities dumping or plotting the curves.

        Finally, some curves (notably those from DI_VDOS) will have the notion
        of "original" and processed curves. By default, the processed curves are
        displayed, although several methods have an "orig" parameter which can
        be set to True, in order to access the original curves instead.
        """
        from . import _miscimpl
        self.__d = _miscimpl._anyvdos_init( AnyVDOS, lambda x : x.__d,
                                            data, fmt, label )

    def __choice( self, orig ):
        return '_orig' if ( orig and 'dos_orig' in self.__d ) else ''

    def integral( self, *, orig = False ):
        """Get the integral of the VDOS curve. Note that this includes the area
        under the parabola between (0,0) and the first positive grid point. Any
        non-positive grid points are ignored for the purposes of calculating the
        integral."""
        return self.__d['derived']['dos%s_integral'%self.__choice(orig)]

    def egrid( self, *, orig = False ):
        """Access the energy grid values. Note that this is always the full
        energy grid, even if just a tuple "(emin,emax)" was used to construct
        the object.
        """
        choice = self.__choice(orig)
        eg = self.__d['derived'].get('dos%s_expandedegrid'%choice)
        return eg if eg is not None else self.__d['dos%s'%choice][0]

    def dos( self, *, orig = False, norm = True ):
        """Access the tabulated DOS values. Unless norm=False, this will
        actually be the normalised DOS, i.e. the DOS values divided by the
        integral (c.f. the .integral property)."""
        choice = self.__choice(orig)
        dos = self.__d['dos%s'%choice][1]
        return ( dos / self.__d['derived']['dos%s_integral'%choice] ) if norm else dos

    @property
    def has_orig( self ):
        """Whether or not this particular object has both processed and original curves."""
        return 'dos_orig' in self.__d

    @property
    def label( self ):
        """A string label describing the curve."""
        return self.__d['label']

    @property
    def debye_temperature( self ):
        """
        The Debye temperature in kelvin, if any is associated with the
        curve. Otherwise it will be None. This is usually only set for
        DI_VDOSDebye objects, or when a single floating point value (the Debye
        temperature itself) was used to initialise the AnyVDOS object.
        """
        return self.__d.get('debyetemp')




################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

"""

Utilities for customising NCrystal's data sources, for instance by adding
directories with data files to the search path, or adding in-memory files.

"""

#from .exceptions import *
from ._chooks import *
_rawfct = _get_raw_cfcts()
import pathlib as _pathlib

def registerInMemoryFileData(virtual_filename,data):
    """Register in-memory file data. This needs a "filename" and the content of this
       virtual file. After registering such in-memory "files", they can be used
       as file names in cfg strings or MatCfg objects. Registering the same
       filename more than once, will simply override the content.

       As a special case data can specified as "ondisk://<path>",
       which will instead create a virtual alias for an on-disk file.
    """
    if ( isinstance(data,str) and data.startswith('ondisk://')):
        data = 'ondisk://'+str(_pathlib.Path(data[9:]).resolve())
    _rawfct['ncrystal_register_in_mem_file_data'](virtual_filename,data)


def addCustomSearchDirectory(dirpath):
    """Register custom directories to be monitored for data files."""
    _rawfct['ncrystal_add_custom_search_dir'](_str2cstr(str(_pathlib.Path(dirpath).resolve())))

def removeCustomSearchDirectories():
    """Remove all search directories added with addCustomSearchDirectory."""
    _rawfct['ncrystal_remove_custom_search_dirs']()

def removeAllDataSources():
    """Disable all standard data sources, remove all TextData factories as well,
       clear all registered virtual files and custom search directories. Finish
       by calling global clearCaches function ("Ripley: I say we take off and
       nuke the entire site from orbit. It's the only way to be sure.").
    """
    _rawfct['ncrystal_remove_all_data_sources']()

def enableAbsolutePaths( enable = True ):
    """Whether or not absolute file paths are allowed."""
    _rawfct['ncrystal_enable_abspaths'](1 if enable else 0)

def enableRelativePaths( enable = True ):
    """Whether or not paths relative to current working directory are allowed."""
    _rawfct['ncrystal_enable_relpaths'](1 if enable else 0)

def enableStandardSearchPath( enable = True ):
    """Whether or not the standard search path should be searched. This standard
      search path is is by default searched *after* the standard data library,
      and is built by concatenating entries in the NCRYSTAL_DATA_PATH
      environment variables with entries in the compile time definition of the
      same name (in that order). Note that by default the standard search path
      is searched *after* the standard data library.
    """
    _rawfct['ncrystal_enable_stdsearchpath'](1 if enable else 0)

def enableStandardDataLibrary( enable = True, dirpath_override = None ):
    """Whether or not the standard data library shipped with NCrystal should be
       searched.

       Unless NCrystal is configured to have the standard data library embedded
       into the binary at compilation time, the location (directory path) of the
       standard data library is taken from the NCRYSTAL_DATADIR environment
       variable. If the environment variable is not set, the location is taken
       from the compile time definition of the same name. If neither is set, and
       data was not embedded at compilation time, the standard data library will
       be disabled by default and the location must be provided before it can be
       enabled. In all cases, the location can be overridden if explicitly
       provided by the user as the second parameter to this function.
    """
    import ctypes
    d = _str2cstr(str(_pathlib.Path(dirpath_override).resolve())) if dirpath_override else ctypes.cast(None, ctypes.c_char_p)
    _rawfct['ncrystal_enable_stddatalib'](1 if enable else 0, d)

class FileListEntry:
    """Entry in list returned by browseFiles."""
    def __init__(self,*,name,source,factName,priority):
        self.__n = name or None
        self.__f = factName or None
        self.__p = int(priority) if priority.isdigit() else priority
        self.__s = source or None

    @property
    def name(self):
        """The (possibly virtual) filename needed to select this entry"""
        return self.__n

    @property
    def source(self):
        """Description (such as the parent directory in case of on-disk files)"""
        return self.__s

    @property
    def factName(self):
        """Name of the factory delivering entry."""
        return self.__f

    @property
    def priority(self):
        """The priority value of the entry (important in case multiple factories
        delivers content with the the same name). Can be 'Unable',
        'OnlyOnExplicitRequest' or an integer priority value (entries with
        higher values will be preferred).
        """
        return self.__p

    @property
    def fullKey(self):
        """The string '%s::%s'%(self.factName,self.name), which can be used to
           explicitly request this entry without interference from similarly
           named entries in other factories.
        """
        return '%s::%s'%(self.__f,self.__n)

    def __str__(self):
        l=[]
        if self.__n:
            l+=['name=%s'%self.__n]
        if self.__s:
            l+=['source=%s'%self.__s]
        if self.__f:
            l+=['factory=%s'%self.__f]
        l+=['priority=%s'%self.__p]
        return 'FileListEntry(%s)'%(', '.join(l))

    def __lt__(self, other):
        if not isinstance(other, FileListEntry):
            return False
        return ( (self.__f,self.__n,self.__s,self.__p)
                 < (other.__f,other.__n,other.__s,other.__p) )

def browseFiles(dump=False,factory=None):
    """Browse list of available input files (virtual or on-disk). The list is not
       guaranteed to be exhaustive, but will usually include all files in
       supported files in the most obvious locations (the NCrystal data
       directory and other directories of the standard search path, the current
       working directory, virtual files embedded in the NCrystal library or
       registered dynamically.

       Returns a list of FileListEntry objects. If the dump flag is set to True,
       the list will also be printed to stdout in a human readable form.

       Setting factory parameter will only return / print entries from the
       factory of that name.

    """
    res=[]
    def sortkey(e):
        praw = e.priority
        if praw=='Unable':
            p=-2
        elif isinstance(praw,int):
            p=praw
        else:
            assert praw=='OnlyOnExplicitRequest'
            p=-1
        return (-p, e.factName,e.source,e.name)
    for n,s,f,p in _rawfct['ncrystal_get_filelist']():
        res.append( FileListEntry(name=n,source=s,factName=f,priority=p) )
    res.sort(key=sortkey)
    if dump:
        seen_names=set()
        groupfct = lambda e : (e.factName,e.source,e.priority)
        lastgroup = None
        pending=[]
        def print_pending():
            if not pending:
                return
            if factory is not None and lastgroup[0]!=factory:
                pending.clear()
                return
            n=len(pending) - 1
            pending[0] = pending[0]%('%s files'%n if n!=1 else '%s file'%n )
            for line in pending:
                print (line)
            pending.clear()
        for e in res:
            group = groupfct(e)
            if lastgroup != group:
                print_pending()
                lastgroup = group
                pending.append('==> %%s from "%s" (%s, priority=%s):'%group)
            hidden = e.name in seen_names
            seen_names.add(e.name)
            extra=''
            prname=e.name
            if e.priority=='OnlyOnExplicitRequest':
                prname='%s::%s'%(e.factName,e.name)
            elif hidden:
                extra=' <--- Hidden by higher priority entries (select as "%s::%s")'%(e.factName,e.name)
            pending.append(    '    %s%s'%(prname,extra))
        print_pending()
        return #return None in this case, to avoid spurious printouts in an interactive session
    if factory is None:
        return res
    return [e for e in res if e.factName==factory]



"""Module with VDOS-related utilities"""

################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

from .constants import constant_planck as _constant_planck
from .constants import constant_boltzmann as _constant_boltzmann
from .constants import constant_c as _constant_c

vdos_units_2_eV = {
    'eV' : 1.0,
    'meV' : 1e-3,
    'keV' : 1e3,
    'MeV' : 1e6,
    'THz' : _constant_planck*1e12,
    'GHz' : _constant_planck*1e9,
    'MHz' : _constant_planck*1e6,
    'kHz' : _constant_planck*1e3,
    'Hz' : _constant_planck,
    '1/cm' : _constant_planck * _constant_c * 1e-8,# 1e-8 to get c in [cm/s] instead of [Aa/s]
}

def createVDOSDebye( debye_temperature ):
    """Create simplified VDOS according to the Debye model"""
    from ._numpy import _np, _ensure_numpy, _np_linspace
    _ensure_numpy()
    #NB: Must keep function exactly synchronised with createVDOSDebye function
    #in .cc src (although leaving out temperature,boundXS,elementMassAMU args
    #here):
    debye_energy = _constant_boltzmann*debye_temperature
    if 0.5*debye_energy < 1.001e-5:
        from .exceptions import NCBadInput
        raise NCBadInput('Too low Debye temperature')
    vdos_egrid = _np_linspace(0.5*debye_energy,debye_energy,20)
    scale = 1.0 / (debye_energy*debye_energy)
    vdos_density = scale * (vdos_egrid**2)
    #Actual returned egrid should contain only first and last value:
    return (_np.asarray([vdos_egrid[0],vdos_egrid[-1]]) ,vdos_density)

def debyeIsotropicMSD( *, debye_temperature, temperature, mass ):
    """Estimate (isotropic, harmonic) atomic mean-squared-displacement (a.k.a. "U_iso") using the
       Debye Model (eq. 11+12 in R.J. Glauber, Phys. Rev. Vol98 num 6,
       1955). Unit of returned MSD value is Aa^2. Input temperatures should be
       in Kelvin, and input atomic mass should be in amu.
    """
    from ._chooks import _get_raw_cfcts
    return float(_get_raw_cfcts()['ncrystal_debyetemp2msd'](debye_temperature, temperature, mass))

def debyeTempFromIsotropicMSD( *, msd, temperature, mass ):
    """The inverse of debyeIsotropicMSD (implemented via root-finding), allowing to
       get the Debye temperature which will give rise to a given
       mean-squared-displacement (a.k.a. "U_iso").
    """
    from ._chooks import _get_raw_cfcts
    return float(_get_raw_cfcts()['ncrystal_msd2debyetemp'](msd, temperature, mass))

def analyseVDOS(emin,emax,density,temperature,atom_mass_amu):
    """Analyse VDOS curve to extract mean-squared-displacements, Debye temperature,
    effective temperature, gamma0 and integral. Input VDOS must be defined via
    an array of density values, over an equidistant energy grid over [emin,emax]
    (in eV). Additionally, it is required that emin>0, and a parabolic trend
    towards (0,0) will be assumed for energies in [0,emin]. Units are kelvin and
    eV where appropriate.
    """
    from ._chooks import _get_raw_cfcts
    return _get_raw_cfcts()['nc_vdoseval'](emin,emax,density,temperature,atom_mass_amu)

def extractGn( vdos, n, mass_amu, temperature, scatxs = 1.0, expand_egrid = True ):
    """Extract Sjolander's Gn function of order n."""
    assert 1 <= n <= 99999
    from .misc import AnyVDOS
    v = AnyVDOS(vdos)
    from ._chooks import _get_raw_cfcts
    emin, emax, Gn =  _get_raw_cfcts()['raw_vdos2gn'](v.egrid(),v.dos(),scatxs, mass_amu, temperature, int(n) )
    if not expand_egrid:
        return (emin,emax),Gn
    else:
        from ._numpy import _ensure_numpy, _np_linspace
        _ensure_numpy()
        return _np_linspace( emin, emax, len(Gn) ), Gn

def extractKnl( vdos, mass_amu, temperature, vdoslux = 3, scatxs = 1.0,
                order_weight_fct = None, plot = False, **plotkwargs ):
    """Expand the VDOS to a scattering kernel, based on the provided
    parameters.

    If provided, the order_weight_function can be used to assign a weight to
    each order of the expansion (e.g. 0.0 to remove the contribution of that
    order). It must be a function taking a single parameter n (the order), and
    return a floating point value (the weight).

    If plot=True, the extracted kernel will be plotted with the
    NCrystal.plot.plot_knl function.
    """
    from .misc import AnyVDOS
    v = AnyVDOS(vdos)
    from ._chooks import _get_raw_cfcts
    a,b,sab = _get_raw_cfcts()['raw_vdos2knl'](v.egrid(),v.dos(),scatxs, mass_amu, temperature, vdoslux, order_weight_fct )
    k = dict(alpha=a, beta=b, sab=sab, mass_amu=mass_amu,temperature=temperature,scatxs=scatxs)
    if plot:
        from .plot import plot_knl
        plot_knl( k, **plotkwargs )
    return k


class PhononDOSAnalyser:

    """Immutable class which reads and facilitates interpretation of phonon DOS information.

    The input data format (fmt) can either be "raw", which is a list of DOS
    (label,egrid,density). It can also be an NCrystal.misc.AnyVDOS,
    NCrystal.Info.DI_VDOS or NCrystal.Info.DI_VDOSDebye object, or a list of
    such. Or it can be "quantumespresso", pointing to a matdyn.dos file produced
    by QuantumEspresso, expecting a format like (leftmost column is the
    frequency, the next column is a combined DOS and the rest are partial DOS's
    for the atoms in the same order as input into QE):

     # Frequency[cm^-1] DOS PDOS
     -2.2774053913E+02  0.0000000000E+00  0.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00
     -2.2674053913E+02  3.3672834745E-05  1.0609E-05  1.0446E-05  6.2610E-06  6.3576E-06
     -2.2574053913E+02  7.5113401337E-05  2.3699E-05  2.3292E-05  1.3952E-05  1.4171E-05
     -2.2474053913E+02  1.1552338186E-04  3.6486E-05  3.5824E-05  2.1434E-05  2.1780E-05
     ...

    The partial DOS columns (the 4 rightmost columns in the example above) will
    be labelled "pdos_0", "pdos_1", etc. After construction, the .plot(..)
    method is then typically used to inspect the data, and the labels can be
    merged, dropped or changed using the methods .drop(..), .merge(..) or
    .update_label(..).

    Next, one might way to apply a lower threshold on the curves if needed with
    .apply_cutoff(..), after choosing an appropriate threshold value by look at
    the plots produced by plot_cutoff_effects and plot_cutoff_effects_on_xsects.

    A word warning: The PhononDOSAnalyser is immutable, hence this code does not
    do anything (mydosana is an PhononDOSAnalyser object):

      mydosana.apply_cutoff( 0.015 )

    Instead one must do:

      mydosana_with_cutoff = mydosana.apply_cutoff()

    Or simply (if you don't care to keep the old un-edited object around):

      mydosana = mydosana.apply_cutoff()

    In addition to setting a threshold with .apply_cutoff, for output files
    intended to have a high quality but smaller footprint (e.g. files destined
    for the NCrystal stdlib) one might also consider applying a regularisation,
    i.e. ensuring that the energy grid is linearly spaced and all grid points
    being a multiple of the binwidth (so that 0.0 would be a grid point, if the
    grid would have been extended downwards).

    Once the phonon curves are as desired, one most likely want to use the
    .apply_to(..) method to apply all or selected VDOS curves to an
    NCMATComposer instance (which one might then for instance use to write the
    data to an NCMAT file).

    Note that many methods on this class can optionally take a list of labels or
    indices. If so, the operation will only affect those DOS curves.

    """

    def __init__(self, data, fmt = None ):
        """Initialise from data (see class description)."""
        from .exceptions import NCBadInput

        if isinstance(data,tuple) and len(data)==2 and data[0]=='__internal_state__' and isinstance(data[1],dict):
            import copy
            self.__d = copy.deepcopy( data[1] )
            return

        if fmt is None:
            from .misc import AnyVDOS
            from .core import Info
            _is_anyvdos = lambda x : any( isinstance(data,e) for e in (AnyVDOS,Info.DI_VDOS,Info.DI_VDOSDebye) )
            if _is_anyvdos( data ):
                fmt = 'anyvdos'
            elif data and hasattr(data,'__len__') and all( _is_anyvdos(e) for e in data ):
                fmt = 'anyvdos_list'
            elif data and hasattr(data,'__len__') and all( ( hasattr(e,'__len__') and len(e)==3 ) for e in data ):
                fmt = 'raw'
            else:
                fmt = 'quantumespresso'

        def _extract_from_anyvdos(x):
            v = AnyVDOS( x )
            return v.label,v.egrid(),v.dos()

        if fmt == 'anyvdos':
            doslist = [_extract_from_anyvdos(data)]
        elif fmt == 'anyvdos_list':
            doslist = [_extract_from_anyvdos(e) for e in data]
        elif fmt == 'quantumespresso':
            from .misc import AnyTextData
            td = AnyTextData( data )
            doslist = _read_quantumespresso( td.content )
            name = td.name
        elif fmt == 'raw':
            import copy
            doslist = copy.deepcopy( data )
            name = 'raw data'
        else:
            raise NCBadInput('Invalid input format. Currently only "raw", "anyvdos", "anyvdos_list", and "quantumespresso" are supported.')

        _dl = []
        from ._numpy import _ensure_numpy, _np
        _ensure_numpy()

        nminpts = 5
        for lbl,egrid,density in doslist:
            egrid = _np.asfarray(egrid).copy()
            if len(egrid) < nminpts:
                raise NCBadInput(f'DOS egrid has too few points (at least {nminpts} required).')
            if len(density) != len(egrid):
                raise NCBadInput('DOS egrid and density arrays have different lengths.')
            _is_grid = lambda a : _np.all(a[:-1] < a[1:])#stackoverflow question 47004506 but with < instead of <=
            if not _is_grid(egrid):
                raise NCBadInput('DOS egrid does not consist of increasing unique values')
            if _np.isinf(egrid[-1]) or not egrid[-1] > 0.0:
                raise NCBadInput(f'DOS egrid has invalid upper edge value: {egrid[-1]}')

            density = _np.asfarray(density).copy()
            _densmin, _densmax = density.min(), density.max()
            if not _densmax > 0.0:
                raise NCBadInput('Maximum density value is not > 0.0.')
            if _densmin < 0.0:
                if abs(_densmin) < 1e-3*_densmax:
                    density.clip( min = 0.0, out = density )
                    from . import _common as nc_common
                    nc_common.warn('Clipping tiny negative DOS density values to 0.0')
                else:
                    raise NCBadInput('Negative density values observed.')

            #peel off excess zeroes:
            assert nminpts>3
            while len(egrid) > nminpts and egrid[-3]>0.0 and density[-2]==0.0 and density[-1]==0.0:
                egrid = egrid[:-1]
                density = density[:-1]
            while len(egrid) > 5 and density[0]==0.0 and density[1]==0.0:
                egrid = egrid[1:]
                density = density[1:]
            if not density.max() > 0.0:
                raise NCBadInput('Invalid input DOS (non-positive everywhere)')
            _dl.append( ( str(lbl), egrid, density ) )
        doslist = _dl

        self.__d = dict( doslist = doslist, srcname = name )

    @property
    def source_name( self ):
        "Name of source data (such as the file name)."
        return self.__d['srcname']

    @property
    def ndos( self ):
        """Number of DOS curves."""
        return len( self.__d['doslist'])

    @property
    def labels( self ):
        """Available DOS labels."""
        return list(lbl for lbl,_,_ in self.__d['doslist'])

    def dos( self, label_or_idx ):
        """Get DOS by index or label. Returns tuple of two arrays: (egrid,dos)."""
        lbl,egrid,dos = self.__d['doslist'][self.__dosidx( label_or_idx )]
        return egrid,dos

    def drop( self, *dos_labels_or_indices ):
        """
        Return a new instance from which the indicated DOS has been removed.
        DOS can be identified via labels or indices, and it is possible to
        provide a list or tuple of them, to drop multiple
        """
        droplist = self.__dosidxlist( *dos_labels_or_indices )
        if not droplist:
            return self
        o = self.__clone()
        o.__d['doslist'] = [ e for i,e in enumerate(o.__d['doslist']) if not i in droplist ]
        return o

    def update_label( self, old_label_or_idx, newlabel ):
        """Return a new instance in which the label of the indicated DOS curve
        has been updated."""
        idx = self.__dosidx( old_label_or_idx )
        o = self.__clone()
        dl = list(o.__d['doslist'])
        l = list(dl[idx])
        l[0] = str(newlabel)
        dl[idx] = tuple(l)
        o.__d['doslist'] = tuple(dl)
        return o

    def merge( self, *dos_labels_or_indices, weights = None, newlabel = None ):
        """
        Return a new instance from which the indicated DOS curves have been merged.
        If weights is given, it must be a list of weights to to use in this merging.

        If newlabel is not provided, the label of the resulting curve will be
        automatically generated.
        """
        mergelist = self.__dosidxlist( *dos_labels_or_indices )
        if weights and len(weights) != len(mergelist):
            from .exceptions import NCBadInput
            raise NCBadInput('Invalid weights provided (must be list of same length as the number of DOS curves to merge')
        if len(mergelist)==1 and newlabel:
            #simply a label update:
            return self.update_label( mergelist[0], newlabel )
        if len(mergelist)<=1:
            #do nothing:
            return self

        egrids = [(lbl,egrid) for i,(lbl,egrid,dos) in enumerate(self.__d['doslist']) if i in mergelist]
        from ._numpy import _ensure_numpy, _np
        _ensure_numpy()
        newegrid = egrids[0][1].copy()
        for lbl,eg in egrids[1:]:
            if not _np.array_equal(newegrid,eg):
                from .exceptions import NCBadInput
                raise NCBadInput('Can not merge DOS curves with incompatible egrids (problems merging labels "%s" and "%s")'%(egrids[0][0],lbl))

        if weights is None:
            weights = [ 1.0 ] * len(mergelist)
        assert all( w >= 0.0 for w in weights )
        import math
        _wsum = math.fsum(w for w in weights)
        assert _wsum > 0.0
        weights = [ w/_wsum for w in weights ]

        c = None
        for idx,w in zip(mergelist,weights):
            _ = self.dos(idx)[1].copy()
            _ *= w
            if c is None:
                c = _
            else:
                c += _
        assert c is not None

        if newlabel is None:
            newlabel = 'merged(%s)'%(','.join( [lbl for i,(lbl,_,_) in
                                                enumerate(self.__d['doslist'])
                                                if i in mergelist] ))
            while newlabel in self.labels:
                newlabel += '(uniquelabel)'
        o = self.drop( mergelist )
        o.__d['doslist'].append( (newlabel,newegrid,c) )
        return o

    def apply_regularisation( self, target_n, *dos_labels_or_indices, quiet = True ):
        """Returns a new instance where the indicated lower threshold value have been
        "regularised", i.e. put on a linearly spaced grid which if it had been
        extended downwards, would eventually cross the point E=0. It is an error
        to apply it to a curve whose first egrid point is not positive (use the
        .apply_cutoff(..) method first for such curves)."""
        selected = self.__dosidxlist( *dos_labels_or_indices, default_is_all = True )
        if not selected:
            return self
        target_n = int(target_n)
        assert target_n >= 10
        doregfct = lambda _e, _d : _do_regularise( egrid=_e, density=_d, n=target_n, quiet = quiet)
        o = self.__clone()
        oldlist = o.__d['doslist']
        newlist = []
        for idx,(lbl,egrid,dos) in enumerate(oldlist):
            if not idx in selected:
                newlist.append( (lbl,egrid,dos) )
                continue
            assert len(egrid)==len(dos)
            if not egrid[0] > 0.0:
                from .exceptions import NCBadInput
                raise NCBadInput('Can not regularise DOS whose first egrid value '
                                 'is not positive. Use the .apply_cutoff method first.')
            egrid, dos = doregfct( egrid, dos )
            assert len(egrid)==len(dos)
            newlist.append( (lbl,egrid,dos) )
        o.__d['doslist'] = newlist
        return o

    def apply_cutoff( self, threshold, *dos_labels_or_indices ):
        """Returns a new instance where the indicated lower threshold value has been
        applied to all (selected) DOS curves."""
        selected = self.__dosidxlist( *dos_labels_or_indices, default_is_all = True )

        if not selected:
            return self

        threshold = self._parse_threshold( threshold )

        from ._numpy import _ensure_numpy, _np
        _ensure_numpy()

        o = self.__clone()
        oldlist = o.__d['doslist']
        newlist = []
        for idx,(lbl,egrid,dos) in enumerate(oldlist):
            if not idx in selected:
                newlist.append( (lbl,egrid,dos) )
                continue
            assert len(egrid)==len(dos)
            j = _np.argmax( egrid >= threshold )
            egrid,dos = egrid[j:],dos[j:]
            assert egrid[0] >= threshold
            assert len(egrid)==len(dos)
            newlist.append( (lbl,egrid,dos) )
        o.__d['doslist'] = newlist
        return o

    def plot( self, *dos_labels_or_indices,
              do_newfig = True,
              do_show = True,
              do_legend = True,
              logy=None,
              unit = 'eV',
              ymin = None,
              ymax = None,
              xmin = None,
              xmax = None,
             ):
        """Plot contained DOS curves for the selected DOS labels (or indices),
        defaulting to all curves. Set do_show to false to avoid plt.show(), logy
        to default to a semilogy plot, ymin to set a minimum plotrange of the
        y-axis, and finally the unit argument can be used to show the DOS in
        other units (e.g. "THz", "1/cm", "meV", etc.
        """
        self.__plot( *dos_labels_or_indices, do_newfig = do_newfig,
                     do_show = do_show, logy=logy, unit = unit,
                     do_legend=do_legend,
                     ymin = ymin, ymax = ymax,
                     xmin = xmin, xmax = xmax )

    def plot_gn( self, *dos_labels_or_indices, n=1,
                 temperature = 293.15, masses = None,
                 do_newfig = True, do_show = True, do_legend = True,
                 logy=None, unit = 'eV' ):
        """Similar to .plot() but showing the Sjolander Gn function instead. If
           labels are not elements or isotopes, masses must be provided in a
           list (in daltons). It is an error to try to plot a Gn function for a
           vdos curve whose first egrid point is not positive (i.e. you must
           first .apply_cutoff(..)).
        """
        selected = self.__dosidxlist( *dos_labels_or_indices, default_is_all = True )
        self.__plot( *selected, do_newfig = do_newfig,
                     do_show = do_show, logy=logy, unit = unit,
                     do_legend=do_legend,
                     sjolanderGn = self.__sjolanderGn_args( selected = selected,
                                                            n=n,
                                                            masses=masses,
                                                            temperature=temperature) )

    def plot_cutoff_effects( self, thresholds, *dos_labels_or_indices, gn = None,
                             temperature = 293.15, masses = None,
                             regularise_n_values = None, **plot_kwargs ):
        """Plot the effect of the listed thresholds on the DOS curves for the
        selected DOS labels (or indices), defaulting to all curves. Any
        plot_kwargs will be used as on the .plot() method.

        To see the effect on Sjolander's Gn curves instead, supply the gn
        keyword (i.e. gn=1 to see G1) -- optionally along with temperature and
        masses keywords (cf. .plot_gn(..)).

        """
        selected = self.__dosidxlist( *dos_labels_or_indices, default_is_all = True )

        if gn is not None:
            sjolanderGn = self.__sjolanderGn_args( selected = selected,
                                                   n=gn,
                                                   masses=masses,
                                                   temperature=temperature)
            plot_kwargs['sjolanderGn'] = sjolanderGn

        unitname,unitfactor = _parsevdosunit( plot_kwargs.get('unit','eV') )
        do_show = plot_kwargs.get('do_show',True)
        do_newfig = plot_kwargs.get('do_newfig',True)
        color_offset = plot_kwargs.get('color_offset',0)
        plot_kwargs['do_show'] = False
        plot_kwargs['do_newfig'] = False
        do_legend = plot_kwargs.get('do_legend',True)
        do_grid = plot_kwargs.get('do_grid',True)
        plot_kwargs['do_grid'] = False
        plot_kwargs['do_legend'] = False

        from .plot import _import_matplotlib_plt
        plt = _import_matplotlib_plt()
        if do_newfig:
            plt.figure()

        plot_thresholds = [e for e in thresholds]
        if gn is None:
            #Can not plot Gn without a cutoff
            plot_thresholds = [None] + plot_thresholds
        reg_n_vals = [None] + ( [ e for e in regularise_n_values] if regularise_n_values else [] )
        for t in plot_thresholds:
            for regn in reg_n_vals:
                plot_kwargs['color_offset'] = color_offset
                if t is None:
                    if regn is None:
                        lblcomments=['orig']
                    else:
                        lblcomments = [f'npts={regn}']
                else:
                    t_parsed = self._parse_threshold( t )
                    lblcomments = [f'cut@{t_parsed/unitfactor:g}{unitname}']
                    if not regn is None:
                        lblcomments+=[f'npts={regn}']
                lblcomment = ','.join(lblcomments)
                plot_kwargs['labelfct'] = lambda lbl : f'{lbl} ({lblcomment})'
                color_offset += len(selected)
                o = self if t is None else self.apply_cutoff( t, *selected )
                if not regn is None:
                    if t is None:
                        #Only go ahead in this case, if all selected egrids already have a cutoff:
                        if not all( self.dos(idx)[0][0]>0.0 for idx in selected ):
                            continue
                    o = o.apply_regularisation( regn, *selected )
                o.__plot(*selected,**plot_kwargs)

        from .plot import _plt_final
        _plt_final(do_grid,do_legend,do_show)

    def plot_cutoff_effects_on_xsects( self, ncmatcomposer, thresholds, cfg_params = None,
                                       *dos_labels_or_indices, lblmap = None, **plot_kwargs ):
        from .ncmat import NCMATComposer
        assert isinstance(ncmatcomposer,NCMATComposer), ( "First argument in call to .plot_cutoff_"
                                                          "effects_on_xsects(..) must be an NCMATComposer object" )
        selected = self.__dosidxlist( *dos_labels_or_indices, default_is_all = True )
        unitname,unitfactor = _parsevdosunit( plot_kwargs.get('unit','eV') )
        do_show = plot_kwargs.get('do_show',True)
        do_newfig = plot_kwargs.get('do_newfig',True)
        do_grid = plot_kwargs.get('do_grid',True)
        do_legend = plot_kwargs.get('do_legend',True)
        plot_kwargs['scatter_breakdown'] = False
        plot_kwargs['show_scattering'] = True
        plot_kwargs['show_absorption'] = False
        plot_kwargs['do_newfig'] = False
        plot_kwargs['do_show'] = False
        plot_kwargs['do_grid'] = False
        plot_kwargs['do_legend'] = False
        plot_kwargs['cfg_params'] = cfg_params
        color = plot_kwargs.get('color')
        lblmap = self.__determine_lblmap( selected, ncmatcomposer, lblmap = lblmap, warn = True )
        lbls = list(sorted(lblmap.keys()))
        if not lbls:
            return
        from .plot import _import_matplotlib_plt
        plt = _import_matplotlib_plt()
        if do_newfig:
            plt.figure()
        colorder = self.__colorder()
        for iplot, t in enumerate([e for e in thresholds]):
            t = self._parse_threshold( t )
            c = ncmatcomposer.clone()
            o = self.apply_cutoff( t, lbls )
            thr_description = f'cut@{t/unitfactor:g}{unitname}'
            for lbl in lbls:
                if not lbl in lblmap:
                    from . import _common as nc_common
                    nc_common.warn('Not using PhononDOSAnalyser label "{lbl}" in plot.')
                    continue
                c.set_dyninfo_vdos( lblmap[lbl], comment = 'From PhononDOSAnalyser', **o.get_dyninfo_args(lbl) )
            if not color:
                plot_kwargs['color'] = colorder[iplot%len(colorder)]
            plot_kwargs['labelfct'] = lambda x : thr_description
            c.plot_xsect( **plot_kwargs )

        if do_legend:
            plt.legend()
        if do_grid:
            plt.grid()
        t = 'DOS cutoff effect'
        if cfg_params:
            t += ' (%s)'%cfg_params.strip()
        plt.title(t)
        if do_show:
            plt.show()

    def apply_to( self, ncmatcomposer, *dos_labels_or_indices, lblmap = None, warn = True, cutoff = None ):
        """Apply DOS curves to NCMATComposer objects, resulting in updates to
        the relevant dyninfo sections. If lblmap is not given, the
        determine_mapping_to_composer_labels() method is used to infer one
        automatically.

        If the cutoff value is provided, it will be applied to all DOS curves
        first. Otherwise, a warning will be emitted and an ad-hoc threshold will
        be supplied to any curves whose initial frequency point is not positive.

        """
        from .ncmat import NCMATComposer
        assert isinstance(ncmatcomposer,NCMATComposer), ( "First argument in call to .apply"
                                                          "(..) must be an NCMATComposer object" )

        selected = self.__dosidxlist( *dos_labels_or_indices, default_is_all = True )
        if not selected:
            return

        from . import _common as nc_common
        warnfct = nc_common.warn if warn else (lambda s : None)

        lblmap = self.__determine_lblmap( selected, ncmatcomposer, lblmap = lblmap, warn = warn )
        cutoff = ( self._parse_threshold( cutoff ) if cutoff is not None else None ) or 0.0
        _access_egrid = lambda idx : self.__d['doslist'][idx][1]
        selected_needscutoff = [ idx for idx in selected if not _access_egrid(idx)[0] > cutoff ]
        if selected_needscutoff and not cutoff:
            #must autodetermine a suitable cutoff. We take it as 1% of the maximum egrid value:
            cutoff = 0.03 * max( _access_egrid(idx)[-1] for idx in selected_needscutoff )
            warnfct(f'Applying an ad-hoc lower DOS egrid cutoff of {cutoff:g}eV since the cutoff parameter'
                    ' was not provided. For important work, it is recommended to instead select one'
                    ' explicitly (probably after investigating - for instance with the .plot_*() methods).')

        o = self.apply_cutoff( cutoff, *selected_needscutoff ) if (cutoff and selected_needscutoff) else self
        if not lblmap:
            warnfct('Not applying any DOS curves to NCMATComposer object'
                    ' (perhaps try again with the lblmap argument).')
            return
        for srclbl, tgtlbl in sorted(lblmap.items()):
            diargs = o.get_dyninfo_args(srclbl)
            assert diargs['vdos_egrid'][0] > 0.0
            ncmatcomposer.set_dyninfo_vdos( tgtlbl,
                                            comment = f'From PhononDOSAnalyser ("{self.source_name}", atom with label "{srclbl}")',
                                            **diargs )

    def is_regular( self, *dos_labels_or_indices ):
        """A regular DOS has an equidistant energy grid whose first point, e0,
           is positive and a binwidth which divides e0 an even number of times."""
        selected = self.__dosidxlist( *dos_labels_or_indices, default_is_all = True )
        if not selected:
            return True
        return all( _vdos_egrid_is_regular(self.__d['doslist'][idx][1]) for idx in selected )

    def requires_cutoff( self, *dos_labels_or_indices ):
        """Whether or not any of the selected curves has an initial energy grid
           point which is not positive."""
        selected = self.__dosidxlist( *dos_labels_or_indices, default_is_all = True )
        return any( (not (self.__d['doslist'][idx][1][0]>0.0) ) for idx in selected )

    def get_dyninfo_args( self, label_or_idx ):
        """returns a dict with 'vdos_egrid' and 'vdos' keys, suitable for usage
           when calling NCMATComposer.set_dyninfo_vdos(..)"""
        lbl,egrid,dos = self.__d['doslist'][self.__dosidx( label_or_idx )]
        return dict( vdos_egrid = egrid, vdos = dos )

    def determine_mapping_to_composer_labels( self, ncmatcomposer, warn = True ):
        """Try to determine a mapping between labels in this object and an
        NCMATComposer object. The mapping will either be based on Z-values or
        the actual label names themselves.

        Unless warn=False, warnings will be emitted when a label could not be mapped.

        """
        from .ncmat import NCMATComposer
        assert isinstance(ncmatcomposer,NCMATComposer), ( "First argument in call to .determine"
                                                          "_mapping_to_composer_labels(..) must be an NCMATComposer object" )
        #First check if any labels can be mapped based on Z-values:
        from . import _common as nc_common
        from . import atomdata as nc_atomdata
        warnfct = nc_common.warn if warn else (lambda s : None)
        lbl2z = {}
        self_labels = self.labels
        for lbl in self_labels:
            elemiso = nc_common.check_elem_or_isotope_marker( lbl )
            if elemiso:
                lbl2z[lbl] = nc_atomdata.elementNameToZValue(elemiso,allow_isotopes=True)
        lbl2tgt_zbased = {}
        for z in set(lbl2z.values()):
            lbls = set( _lbl for _lbl,_z in lbl2z.items() if _z==z )
            if len(lbls) == 1:
                lbl = lbls.pop()
                tgtlbl = ncmatcomposer.find_label( z, allow_multi = False )
                if tgtlbl:
                    lbl2tgt_zbased[lbl] = tgtlbl

        #Now, see if we can map based on having the same label in the two objects:
        common_lbls = set(self_labels).intersection(set( ncmatcomposer.get_labels() ))

        res = {}
        for lbl in self_labels:
            tgt_z = lbl2tgt_zbased.get(lbl)
            tgt_lblname = lbl if lbl in common_lbls else None
            if tgt_z and tgt_lblname and tgt_z != tgt_lblname:
                warnfct( f'Could not map atom with PhononDOSAnalyser label "{lbl}" '
                         'unambiguously to NCMATComposer (Z value implies mapping to'
                         f' "{tgt_z}" but the label name implies the target label "{tgt_lblname}")')
            else:
                tgtlbl = tgt_z or tgt_lblname
                if tgtlbl:
                    res[lbl] = tgtlbl
                else:
                    warnfct( f'Could not map atom with PhononDOSAnalyser label "{lbl}" '
                             'unambiguously to NCMATComposer label' )
        return res

    def __dosidx( self, label_or_idx ):
        import numbers
        l = self.__d['doslist']
        if isinstance(label_or_idx,numbers.Integral):
            i = int( label_or_idx )
            if not 0 <= i < len(l):
                from .exceptions import NCBadInput
                raise NCBadInput('Index out of range (%i is not in range 0..%i'%(i,len(l)-1))
            return i
        _ = [i for i,(lbl,_,_) in enumerate(l) if lbl == label_or_idx ]
        if not _:
            from .exceptions import NCBadInput
            _ = '","'.join(lbl for lbl,_,_ in l)
            raise NCBadInput('Invalid label "%s" (available labels are "%s")'%(label_or_idx,_))
        assert len(_) == 1
        return _[0]

    def __dosidxlist( self,  *dos_labels_or_indices, default_is_all = False ):
        if not dos_labels_or_indices:
            return list( range( self.ndos ) ) if default_is_all else []
        res = []
        for l in dos_labels_or_indices:
            if not hasattr(l,'__len__') or hasattr(l,'startswith'):
                #single item
                res.append( self.__dosidx( l ) )
            else:
                res += [ self.__dosidx( e ) for e in l ]
        return res

    def __sjolanderGn_args( self, selected, n=1, masses = None, temperature = 293.15 ):
        assert 1<=n<=9999
        assert temperature >= 0.001
        from .exceptions import NCBadInput
        if masses is None:
            masses = []
            from .atomdata import atomDB
            for idx in selected:
                lbl = self.__d['doslist'][idx][0]
                ad = atomDB(lbl,throwOnErrors=False)
                if not ad:
                    raise NCBadInput( f'Can not plot Gn function for label "lbl" which does'
                                      ' not correspond to a known element or isotope. Either'
                                      ' change the label with .update_label(..), or directly'
                                      ' provide masses using the "masses" parameter' )
                masses.append( ad.averageMassAMU() )
        if len(masses) != len(selected):
            raise NCBadInput( 'Invalid number of masses provided' )
        return dict( n = n, masses = masses, temperature=temperature )

    def __colorder( self ):
        from ._common import _palette_Few as _palette
        return [ _palette[e] for e in ['red',
                                       'blue',
                                       'orange',
                                       'green',
                                       'purple',
                                       'yellow',
                                       'pink',
                                       'gray',
                                       ]]

    def __plot( self, *dos_labels_or_indices, do_newfig = True, do_show = True, logy=None, unit = 'eV',
                color_offset = 0, labelfct = None, do_legend=True,do_grid=True, sjolanderGn = None,
                ymin=None, ymax = None, xmin=None, xmax=None ):

        selected = self.__dosidxlist( *dos_labels_or_indices, default_is_all = True )
        gnfcts = {}
        if sjolanderGn is not None:
            for idx, mass in zip(selected,sjolanderGn['masses']):
                lbl, egrid, dos = self.__d['doslist'][idx]
                if not egrid[0]>0.0:
                    from .exceptions import NCBadInput
                    raise NCBadInput('Can not plot Gn functions for DOS curves whose initial grid'
                                     ' point is not at a positive value (use apply_cutoff(..) to rectify')
                gnfcts[idx] = extractGn( vdos = (egrid,dos),
                                         n = sjolanderGn['n'],
                                         mass_amu = mass,
                                         temperature = sjolanderGn['temperature'] )



        unitname,unitfactor = _parsevdosunit( unit )
        from .plot import _import_matplotlib_plt
        plt = _import_matplotlib_plt()
        if do_newfig:
            plt.figure()
        from ._numpy import _ensure_numpy, _np_linspace

        colorder = self.__colorder()
        for idx,(lbl, egrid, dos) in enumerate( self.__d['doslist'] ):
            if not idx in selected:
                continue
            gnfct = gnfcts.get(idx)
            color = colorder[(color_offset+idx)%len(colorder)]
            actual_lbl = lbl if labelfct is None else labelfct(lbl)
            if gnfct:
                plt.plot( gnfct[0]/unitfactor, gnfct[1],
                          label = actual_lbl,
                          color = color )
            else:
                plt.plot( egrid/unitfactor, dos,
                          label = actual_lbl,
                          color = color )
                if egrid[0]>0.0:
                    _k = dos[0] / egrid[0]**2
                    _ensure_numpy()
                    _x = _np_linspace(0.0, egrid[0], 2000+2)[1:-1]
                    plt.plot( _x/unitfactor, _k*(_x**2),ls=':',color=color)

        plt.xlabel('Frequency (%s)'%unitname)
        if sjolanderGn is not None:
            plt.ylabel('G%i (arbitrary scale)'%sjolanderGn['n'])
        else:
            plt.ylabel('DOS (arbitrary scale)')
        if ymin is not None or ymax is not None:
            plt.ylim(ymin,ymax)
        if xmin is not None or xmax is not None:
            plt.xlim(xmin,xmax)
        from .plot import _plt_final
        _plt_final(do_grid,do_legend,do_show,logy=logy)

    def __clone( self ):
        return PhononDOSAnalyser( ('__internal_state__',self.__d) )

    def _parse_threshold( self, value ):
        import numbers
        from ._common import _decodeflt
        from .exceptions import NCBadInput
        def impl(x):
            if isinstance(x,numbers.Real):
                return float(x)
            if isinstance(x,str):
                x=x.strip()
                v = None
                for un,unval in vdos_units_2_eV.items():
                    if x.endswith(un):
                        v = _decodeflt(x[:-len(un)].strip())
                        if v is not None:
                            return v *unval
                if v is None:
                    raise NCBadInput(f'Invalid threshold string: {x}')
            if hasattr(x,'__len__') and len(x)==2 and isinstance(x[1],str):
                _,unitfactor = _parsevdosunit( x[1] )
                v = _decodeflt( x[0] )
                if v is None:
                    raise NCBadInput(f'Invalid threshold value: {x[0]}')
                return v*unitfactor
            raise NCBadInput(f'Invalid threshold: {x}')

        v = impl(value)
        if not v>0.0 or not v < 1e6:
            raise NCBadInput(f'Invalid threshold value (out of range): {v:g}')
        return v

    def __determine_lblmap( self, selected, ncmatcomposer, lblmap = None, warn = True ):
        from .exceptions import NCBadInput
        if lblmap is None:
            lblmap = self.determine_mapping_to_composer_labels( ncmatcomposer, warn = warn )
        else:
            #use provided lblmap, but with a few sanity checks:
            composer_lbls = ncmatcomposer.get_labels()
            missing = set(lblmap.values()).difference(composer_lbls)
            fmtlabellist = lambda l : ( '"%s"'%('", "'.join(l)) if l else '' )
            if missing:
                raise NCBadInput('Some values in lblmap are not present in provided'
                                 f' NCMATComposer object: {fmtlabellist(missing)} (the'
                                 f' following labels are available: {fmtlabellist(composer_lbls)})')
            missing = set(lblmap.keys()).difference(self.labels)
            if missing:
                raise NCBadInput('Some keys in lblmap are not actual labels in PhononDOSAnalyser'
                                 f' object: {fmtlabellist(missing)} (the'
                                 f' following labels are available: {fmtlabellist(self.labels)})')
            lblmap = dict( lblmap.items() )
        res = {}
        for k,v in lblmap.items():
            idx = self.__dosidx(k)
            if idx in selected:
                res[k] = v
        return res


def _read_quantumespresso( raw_text_data ):
    assert isinstance(raw_text_data,str)
    _data_lines = raw_text_data.splitlines()

    def _get_header():
        out=[]
        for l in _data_lines:
            l = l.strip()
            if not l:
                continue
            if not l.startswith('#'):
                break
            out.append( ' '.join(l[1:].split()) )
        return out

    hdr = _get_header()
    _expected_hdr = 'Frequency[cm^-1] DOS PDOS'
    if not hdr or not _expected_hdr in hdr:
        from .exceptions import NCBadInput
        raise NCBadInput('Invalid input format. Did not find expected header line "# %s"'%_expected_hdr)
    from ._numpy import _ensure_numpy, _np
    _ensure_numpy()

    rawdata = _np.loadtxt( _data_lines )
    assert len( rawdata.shape ) == 2
    nrows, ncols = rawdata.shape if len( rawdata.shape ) == 2 else ( 0, 0 )
    if not nrows >= 10 or not ncols >= 3:
        from .exceptions import NCBadInput
        raise NCBadInput('Invalid input format. Expected at least 3 columns and 10 rows')
    npdos = ncols - 2

    _get_col = lambda icol : rawdata[:,icol]
    egrid = vdos_units_2_eV['1/cm'] * _get_col( 0 )

    doslist  = [ ( 'combined_dos', egrid, _get_col( 1 ) ) ]
    doslist += [ ('pdos%i'%ipdos, egrid, _get_col(2+ipdos) ) for ipdos in range(npdos)]
    return doslist

def _vdos_egrid_is_regular( egrid ):
    if not ( egrid[0] > 0.0 ):
        return False
    from ._common import _grid_is_linspace
    if not _grid_is_linspace( egrid ):
        return False
    bw = ( egrid[-1]-egrid[0] ) / ( len(egrid) - 1 )
    x = egrid[0] / bw
    #regular if x is very near an integer >= 1
    nx = round(x)
    return nx >= 1 and abs(x-nx) < 1e-4

def _do_regularise( egrid, density, n, quiet = False):
    #Regularisation function adapted from ncrystal_vdos2ncmat

    if n >= len(egrid) and _vdos_egrid_is_regular(egrid):
        #Already fine, won't get better just by wasting more points!
        return egrid,density

    from ._numpy import _ensure_numpy, _np, _np_linspace
    _ensure_numpy()

    egrid = _np.asfarray(egrid)
    density = _np.asfarray(density)
    assert len(egrid)==len(density)
    assert len(egrid) >= 2
    assert egrid[0] > 0.0

    emin,emax=egrid[0],egrid[-1]
    if quiet:
        nc_print = lambda *a,**kw : None
    else:
        from ._common import print as nc_print
    nc_print('old range',emin,emax)
    THZ = vdos_units_2_eV['THz']
    nc_print('old range [THz]',emin/THZ,emax/THZ)

    for k in range(1,1000000000):
        #k is number of bins below emin, an integral number by definition in a regularised grid.
        binwidth = emin/k
        nbins=int(_np.floor((emax-emin)/binwidth))+1
        eps = (emin+nbins*binwidth)-emax
        assert eps>=0.0
        if nbins+1 >= n:
            break
    n=nbins+1
    binwidth = emin/k
    new_emax = emin + (n-1) * binwidth
    if abs( (new_emax-binwidth) - emax ) < 1e-3*binwidth:
        nbins -= 1
        n -= 1
        new_emax -= binwidth
    nc_print(f" ==> Choosing regular grid with n={n} pts from emin={emin} to emax={new_emax} ({new_emax-emax} beyond old emax)")
    def retry():
        from ._common import warn
        nnew = n+100
        warn('Something went wrong in DOS regularisation with n={n}. Retrying with n={nnew}.')
        return  _do_regularise(egrid,density,n = nnew, quiet = quiet )
    if not ( new_emax >= emax-binwidth*1.001e-3 ):
        return retry()
    if not ( new_emax >= emax-binwidth*1.001e-3 ):
        return retry()
    new_egrid = _np_linspace(emin,new_emax,n)
    test=new_egrid[0] / ( (new_egrid[-1]-new_egrid[0])/(len(new_egrid)-1) )
    if not abs(round(test)-test)<1e-6:
        return retry()
    new_density = _np.interp(new_egrid,egrid,density, left=0.0, right=0.0)
    nc_print(f'last density values in new grid: ',new_density[-5:])
    return new_egrid,new_density

def _parsevdosunit( name ):
    unitfactor = vdos_units_2_eV.get(name or 'eV')
    if unitfactor:
        return name, unitfactor
    from .exceptions import NCBadInput
    _ = '","'.join(u for u in sorted(vdos_units_2_eV.keys()))
    raise NCBadInput('Invalid frequency unit "%s" (must be one of "%s")'%(name,_))



"""

Implementation of the built-in unit tests.

"""

################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

__all__ = ['test','test_cmdline','test_cmake','test_extra','test_all']
from ._common import print as _nc_print

def test( verbose = False ):
    """Quick test that NCrystal works as expected in the current installation."""
    _actualtest( verbose = verbose )
    if verbose!='quiet':
        _nc_print("Tests completed succesfully")

def test_cmdline( verbose = False ):
    """Quick test that NCrystal command-line scripts are present and working in
    the current installation."""
    _actual_test_cmdline( verbose = verbose )
    if verbose!='quiet':
        _nc_print("Tests completed succesfully")

def test_cmake( verbose = False ):
    """Quick test that the NCrystal installation supports down-stream CMake/C++
    projects (possibly after using ncrystal-config to get CMAKE_PREFIX_PATH and
    (DY)LD_LIBRARY_PATHS setup).
    """
    _actual_test_cmake( verbose = verbose, ignore_if_absent = False )
    if verbose!='quiet':
        _nc_print("Tests completed succesfully")

def test_extra( verbose = False ):
    """Run both test() and test_cmdline(). If CMake is present, also run test_cmake()."""
    _actualtest( verbose = verbose )
    _actual_test_cmdline( verbose = verbose )
    _actual_test_cmake( verbose = verbose, ignore_if_absent = True )
    if verbose!='quiet':
        _nc_print("Tests completed succesfully")

def test_all( verbose = False ):
    """Run both test(), test_cmdline(), and test_cmake()."""
    _actualtest( verbose = verbose )
    _actual_test_cmdline( verbose = verbose )
    _actual_test_cmake( verbose = verbose, ignore_if_absent = False )
    if verbose!='quiet':
        _nc_print("Tests completed succesfully")

def _get_prfct( verbose ):
    if verbose and verbose != 'quiet':
        return lambda *a, **kw : _nc_print('::NCrystalTest::',*a,**kw)
    else:
        return lambda *a, **kw : None

def _actualtest( verbose ):
    prfct = _get_prfct( verbose )
    prfct('starting standard Python-API testing')
    try:
        import numpy as _np
    except ImportError:
        _np = None
    from . import core as NC
    from .constants import wl2ekin
    from .plugins import hasFactory

    def require(b):
        if not b:
            raise RuntimeError('check failed')
    def flteq(a,b,rtol=1.0e-6,atol=1.0e-6):
        return abs(a-b) <= 0.5 * rtol * (abs(a) + abs(b)) + atol
    def require_flteq(a,b):
        if not flteq(a,b):
            raise RuntimeError('check failed (%.16g != %.16g, diff %g)'%(a,b,a-b))
        return True
    require(hasFactory('stdncmat'))
    from . import _common as nc_common
    require( nc_common.prettyFmtValue(0.25) == '1/4' )
    require( nc_common.prettyFmtValue(0.25 + 1e-12) != '1/4' )
    require( nc_common.prettyFmtValue(0.25 + 1e-15) == '1/4' )

    _cfgstr='stdlib::Al_sg225.ncmat;dcutoff=1.4'
    prfct(f'Trying to createInfo("{_cfgstr}")')
    al = NC.createInfo(_cfgstr)
    prfct(f'Verifying loaded Info object')
    require(al.hasTemperature() and require_flteq(al.getTemperature(),293.15))
    require_flteq(al.getXSectFree(),1.39667)
    require_flteq(al.getXSectAbsorption(),0.231)
    require_flteq(al.getDensity(),2.69864547673)
    require_flteq(al.getNumberDensity(),0.06023238256131625)
    require(al.hasDebyeTemperature())

    require(al.hasStructureInfo())
    si=al.getStructureInfo()
    require( si['spacegroup'] == 225 )
    require_flteq(si['a'],4.04958)
    require_flteq(si['b'],4.04958)
    require_flteq(si['c'],4.04958)
    require( si['alpha'] == 90.0 )
    require( si['beta'] == 90.0 )
    require( si['gamma'] == 90.0 )
    require( si['n_atoms'] == 4 )
    require_flteq(si['volume'],66.4094599932)
    require( al.hasHKLInfo() )
    require( al.nHKL() == 3 )
    require_flteq(al.hklDLower(),1.4)
    require( al.hklDUpper() > 1e36 )
    expected_hkl = { 0  : (1, 1, 1, 8, 2.3380261031049243, 1.7731590373262052),
                     1  : (2, 0, 0, 6, 2.02479, 1.7317882793764163),
                     2  : (2, 2, 0, 12, 1.4317427394787092, 1.5757351418107877) }
    for idx,hkl in enumerate(al.hklList()):
        h,k,l,mult,dsp,fsq = hkl
        require(idx<len(expected_hkl))
        e = expected_hkl[idx]
        require( list(e)[0:4] == [h,k,l,mult] )
        require_flteq(dsp, e[4])
        require_flteq(fsq, e[5])

    _cfgstr2='stdlib::Al_sg225.ncmat;dcutoff=1.4;incoh_elas=0;inelas=0'
    prfct(f'Trying to createScatter("{_cfgstr2}")')
    #We do all createScatter... here with independent RNG, for reproducibility
    #and to avoid consuming random numbers from other streams.
    alpc = NC.createScatterIndependentRNG(_cfgstr2)
    prfct(f'Verifying loaded Scatter object')
    require( alpc.name == 'PCBragg' )
    require( isinstance(alpc.name,str) )
    require( alpc.refCount() in (1,2) and type(alpc.refCount()) == int )
    require( alpc.isNonOriented() )
    #_nc_print(alpc.xsect(wl=4.0))
    require_flteq(1.632435821586171,alpc.crossSectionIsotropic(wl2ekin(4.0)) )
    require_flteq(1.632435821586171,alpc.crossSection(wl2ekin(4.0),(1,0,0)))
    require( alpc.crossSectionIsotropic(wl2ekin(5.0)) == 0.0 )

    require( alpc.rngSupportsStateManipulation() )
    require(alpc.getRNGState()=='a79fd777407ba03b3d9d242b2b2a2e58b067bd44')

    alpc.setRNGState('deadbeefdeadbeefdeadbeefdeadbeefb067bd44')
    require(alpc.getRNGState()=='deadbeefdeadbeefdeadbeefdeadbeefb067bd44')
    alpc_clone = alpc.clone()
    require(alpc.getRNGState()=='deadbeefdeadbeefdeadbeefdeadbeefb067bd44')
    require(alpc_clone.getRNGState()=='e0fd16d42a2aced7706cffa08536d869b067bd44')
    alpc_clone2 = alpc_clone.clone(for_current_thread=True)
    require(alpc_clone2.getRNGState()=='cc762bb1160a0be514300da860f6d160b067bd44')
    alpc_clone3 = alpc_clone.clone(rng_stream_index = 12345 )
    require(alpc_clone3.getRNGState()=='3a20660a10fd581bd7cddef8fc3f32a2b067bd44')

    #Pick Nickel at 1.2 angstrom, to also both vdos + incoherent-elastic + coherent-elastic:
    _cfgstr3='stdlib::Ni_sg225.ncmat;dcutoff=0.6;vdoslux=2'
    _seed=2543577
    prfct(f'Trying to createScatterIndependentRNG("{_cfgstr3}",{_seed})')
    nipc = NC.createScatterIndependentRNG(_cfgstr3,_seed)
    prfct(f'Verifying loaded Scatter object')
    nipc_testwl = 1.2
    #_nc_print(nipc.xsect(wl=nipc_testwl),nipc.xsect(wl=5.0))
    require_flteq(16.76322537767633,nipc.xsect(wl=nipc_testwl))
    require_flteq(16.76322537767633,nipc.xsect(wl=nipc_testwl,direction=(1,0,0)))
    require_flteq(5.958094744249944,nipc.xsect(wl=5.0))

    require( nipc.name == 'ProcComposition' )

    expected = [ ( 0.056808478892590906, 0.5361444826572666 ),
                 ( 0.056808478892590906, 0.5361444826572666 ),
                 ( 0.056808478892590906, 0.36219866365374176 ),
                 ( 0.056808478892590906, 0.8391056916029316 ),
                 ( 0.032001313096074194, -0.3726211530494784 ),
                 ( 0.056808478892590906, -0.10165685368899147 ),
                 ( 0.056808478892590906, -0.15963879335683306 ),
                 ( 0.056808478892590906, 0.8260541809964751 ),
                 ( 0.07799891342244511, -0.5293689067509328 ),
                 ( 0.05348597239804991, -0.09542895891111686 ),
                 ( 0.056808478892590906, 0.8260541809964751 ),
                 ( 0.04125596596989546, -0.2114086959439411 ),
                 ( 0.056808478892590906, -0.10165685368899147 ),
                 ( 0.056808478892590906, -0.10165685368899147 ),
                 ( 0.056808478892590906, 0.5361444826572666 ),
                 ( 0.056808478892590906, -0.3915665520281999 ),
                 ( 0.056808478892590906, 0.36219866365374176 ),
                 ( 0.05750930296126236, -0.5221485562238027 ),
                 ( 0.056808478892590906, 0.36219866365374176 ),
                 ( 0.0812282226184265, -0.9893430983107131 ),
                 ( 0.056808478892590906, -0.5655123710317247 ),
                 ( 0.058097184485747494, -0.951408724433637 ),
                 ( 0.056808478892590906, 0.3042167239859003 ),
                 ( 0.056808478892590906, 0.7378808571510718 ),
                 ( 0.056808478892590906, -0.10165685368899147 ),
                 ( 0.08045270890565197, -0.8062090812584963 ),
                 ( 0.056808478892590906, -0.5655123710317247 ),
                 ( 0.06930621305459778, 0.0790013056899895 ),
                 ( 0.04019454300337846, -0.9619857378392679 ),
                 ( 0.08983663449895618, -0.5822245509723732 ) ]

    if _np is None:
        ekin,mu=[],[]
        for i in range(30):
            _ekin,_mu=nipc.sampleScatterIsotropic(wl2ekin(nipc_testwl))
            mu += [_mu]
            ekin += [_ekin]
    else:
        ekin,mu = nipc.sampleScatterIsotropic(wl2ekin(nipc_testwl),repeat=30)

    for i in range(len(ekin)):
        #print ( f'    ( {ekin[i]}, {mu[i]} ),');continue
        require_flteq(ekin[i],expected[i][0])
        require_flteq(mu[i],expected[i][1])

    expected = [ ( 0.056808478892590906, (0.07228896531453344, -0.5190173207165885, 0.8517014302500192) ),
                 ( 0.056808478892590906, (-0.9249112255344181, -0.32220112076758217, -0.20180600252850442) ),
                 ( 0.056808478892590906, (-0.15963879335683306, -0.8486615569734178, 0.5042707778277745) ),
                 ( 0.0492224669270449, (-0.9779916301402904, 0.140975569549056, 0.15381241876342955) ),
                 ( 0.056808478892590906, (0.07228896531453344, 0.7905105193171594, -0.6081672667471253) ),
                 ( 0.056808478892590906, (-0.10165685368899147, -0.8869759070713323, -0.4504882066969593) ),
                 ( 0.056808478892590906, (0.07228896531453344, -0.39741541395284924, -0.914787021249449) ),
                 ( 0.056808478892590906, (-0.10165685368899147, -0.9768880366798581, -0.1880309758785167) ),
                 ( 0.025610418737037184, (-0.884783168996826, -0.4657283985847863, 0.015993830422524287) ),
                 ( 0.056808478892590906, (0.8260541809964751, 0.539797243436807, 0.16202909009269678) ),
                 ( 0.07443181306716587, (-0.6036941700256581, -0.06282145095069493, -0.7947369466543514) ),
                 ( 0.056808478892590906, (0.8260541809964751, 0.10854661864786977, 0.5530389874487663) ),
                 ( 0.056808478892590906, (0.5361444826572666, 0.7795115518549294, 0.3238994199452849) ),
                 ( 0.056808478892590906, (0.07228896531453344, 0.746175597107444, 0.6618128767069312) ),
                 ( 0.056808478892590906, (-0.10165685368899147, -0.4247181868490453, 0.8996001033001911) ),
                 ( 0.056808478892590906, (0.5361444826572666, 0.5555760611065321, -0.6355189486093415) ),
                 ( 0.05736918247226906, (-0.17265143322057086, -0.684984059616355, 0.7078052844380157) ),
                 ( 0.056808478892590906, (0.3042167239859003, -0.8706122815482211, -0.3866347631352975) ),
                 ( 0.056808478892590906, (-0.7384733804796917, 0.6322144258925643, -0.23443972789660028) ),
                 ( 0.056808478892590906, (-0.15963879335683306, 0.21525619037302965, -0.9634211063505222) ),
                 ( 0.056808478892590906, (0.41359447569500096, 0.4927058865194684, 0.7656242675514158) ),
                 ( 0.056808478892590906, (0.25796367721315083, 0.48520231047621615, 0.8354839670198411) ),
                 ( 0.056808478892590906, (0.5785005938702705, 0.8104481067271115, -0.09225469740985966) ),
                 ( 0.04320288783696474, (-0.030385860971878235, -0.49547867364837517, 0.8680884651996275) ),
                 ( 0.05428746831317616, (-0.3602629693021255, -0.9063804616575692, 0.22064689364463652) ),
                 ( 0.056808478892590906, (0.36219866365374176, -0.8822186430862216, 0.3008361577978114) ),
                 ( 0.056808478892590906, (0.7680722413286334, 0.5975216576265994, -0.23028873347945303) ),
                 ( 0.056808478892590906, (0.32922859149927786, -0.9426419619170849, 0.0550878042084668) ),
                 ( 0.056808478892590906, (-0.10165685368899147, -0.2489220191768986, -0.9631737706493833) ),
                 ( 0.0670456981700729, (-0.8979842133391931, 0.34668021323231085, 0.2709929562678522) ) ]

    for i in range(30):
        out_ekin,outdir = nipc.sampleScatter(wl2ekin(nipc_testwl),(1.0,0.0,0.0))
        #print ( f'    ( {out_ekin}, {outdir} ),');continue
        require_flteq(out_ekin,expected[i][0])
        require_flteq(outdir[0],expected[i][1][0])
        require_flteq(outdir[1],expected[i][1][1])
        require_flteq(outdir[2],expected[i][1][2])

    _cfgstr4="""stdlib::Ge_sg227.ncmat;dcutoff=0.5;mos=40.0arcsec
                            ;dir1=@crys_hkl:5,1,1@lab:0,0,1
                            ;dir2=@crys_hkl:0,-1,1@lab:0,1,0""".replace(' ','').replace('\n','')
    _seed4 = 3453455
    prfct(f'Trying to createScatterIndependentRNG("{_cfgstr4}",{_seed4})')
    gesc = NC.createScatterIndependentRNG(_cfgstr4,_seed4)
    prfct(f'Verifying loaded Scatter object')
    require_flteq(591.025731949681,gesc.crossSection(wl2ekin(1.540),( 0., 1., 1. )))
    require_flteq(1.666984885615526,gesc.crossSection(wl2ekin(1.540),( 1., 1., 0. )))
    prfct('standard Python-API testing done')


class CallInspector:

    @property
    def name( self ):
        return self.__name

    @property
    def the_real_inspected_object( self ):
        return self.__realobj

    def __init__( self, name, *, realobj = None, subfcts = None, fmtcall = None ):
        self.__name = name
        self.__realobj = realobj
        self.__subfcts = {}
        self.__fmtcall = fmtcall or _fmtcall
        for sf in (subfcts or []):
            if isinstance(sf,str):
                self.__subfcts[sf] = None
            else:
                assert len(sf)==2 and isinstance(sf,tuple)
                self.__subfcts[sf[0]] = sf[1]

    def __call__( self, *a, **kwargs ):
        return self.__getattrimpl('__call__')(*a,**kwargs)

    def __getattr__( self, attrname ):
        return self.__getattrimpl(attrname)

    def __getattrimpl( self, attrname ):
        if not attrname in self.__subfcts:
            raise RuntimeError(f'Not allowed to access attribute {attrname} of {self.__name} objects.')
        sf_kwargs = self.__subfcts[attrname] or {}
        thefmtcall = sf_kwargs.get('fmtcall') or self.__fmtcall
        def wrapper( *args, **kwargs ):
            _n = f'{self.__name}.{attrname}' if attrname!='__call__' else self.__name
            _nc_print(f"CALLING %s"%thefmtcall(_n,args,kwargs))
            res = getattr(self.__realobj,attrname)(*args,**kwargs) if self.__realobj else None
            return CallInspector( name = 'ResultOf[%s(..)]'%_n,
                                  realobj = res, **sf_kwargs ) if sf_kwargs else None
        return wrapper

def _fmtvalue( x ):
    import numbers
    if isinstance(x,numbers.Real) and not isinstance(x,numbers.Integral):
        s='%.3g'%x#only 3 to avoid too many spurious test issues due to FPE irrep.
        return s+'.0' if ( s.isdigit() or s[0]=='-' and s[1:].isdigit() ) else s
    return repr(x)

def _fmtcall(fctname,args=tuple(),kwargs={}):
    import numbers
    def _fmt(a):
        if isinstance(a,numbers.Real):
            return _fmtvalue(a)
        def pruneaddr(s):
            while ' object at 0x' in s:
                _ = s.split(' object at 0x',1)
                while _[1] and _[1][0].isalnum():
                    _[1] = _[1][1:]
                s = ' object at SNIPADDR'.join(_)
            return s
        return 'Object[%s]'%a.name if isinstance(a,CallInspector) else pruneaddr(repr(a))
    l = [ _fmt(a) for a in args ]
    l += [ '%s=%s'%(k,_fmt(v)) for k,v in sorted(kwargs.items()) ]
    a=','.join(l)
    return f'{fctname}({a})'

def _create_pyplot_inspector( pass_calls_to_real_plt ):
    import copy
    if pass_calls_to_real_plt:
        import matplotlib.pyplot as realplt
    else:
        realplt = None
    def shorten( x ):
        if hasattr(x,'shape') and len(x.shape)==2:
            return 'Array(shape=%s,content=%s)'%(x.shape,shorten(x.flatten()))
        if isinstance(x,str) or not hasattr(x,'__len__'):
            return x
        #only 4 digits to guard against annoying test errors:
        if len(x) <= 10:
            return list(_fmtvalue(e) for e in x)
        else:
            return list(_fmtvalue(e) for e in x[0:3])+['...']+list(_fmtvalue(e) for e in x[-3:])
    def _create_shortening_fmtcall( nargs_to_shorten, kwargs_to_shorten = None ):
        def fmtcall_pltplot( name, args, kwargs ):
            plot_args, plot_kwargs = args, kwargs
            if nargs_to_shorten>0:
                plot_args = [ shorten(e) for e in args[0:nargs_to_shorten] ] + list(args[nargs_to_shorten:])
            if any( e in kwargs for e in (kwargs_to_shorten or [])):
                plot_kwargs = copy.deepcopy(kwargs)
                for xy in ('x','y'):
                    if xy in plot_kwargs:
                        plot_kwargs[xy] = shorten(plot_kwargs[xy])
            return _fmtcall(name, plot_args,plot_kwargs)
        return fmtcall_pltplot

    return CallInspector( name = 'plt',
                          realobj = realplt,
                          subfcts = [ 'title','xlabel','ylabel','semilogx','semilogy',
                                      'legend','grid','show','figure',
                                      'savefig','ylim','xlim','colorbar',
                                      'tight_layout','close',
                                      ( 'plot', dict( fmtcall=_create_shortening_fmtcall(2,('x','y')) ) ),
                                      ( 'pcolormesh',dict( subfcts=['set_clim'],
                                                           fmtcall=_create_shortening_fmtcall(3) ) ),
                                     ] )

def _create_pdfpages_inspector( real_pdfpages ):
    return CallInspector( name = 'PdfPages',
                          realobj = real_pdfpages,
                          subfcts = [ ('__call__',dict(subfcts=['savefig','close'])) ] )

def _run_cmd( cmd, env = None ):
    import sys
    import subprocess
    sys.stdout.flush()
    sys.stderr.flush()
    if env and sys.platform == 'darwin':
        #osx's system protection #$%#^ is so damn annoying:
        for name in ['DYLD_LIBRARY_PATH','LD_LIBRARY_PATH']:
            v = env.get(name)
            if v is not None:
                from shlex import quote
                #NB: "export A=B" is posix, not bash, so hopefully OK on all OSX
                #shells:
                cmd = 'export %s=%s && %s'%(name,quote(v),cmd)
    try:
        p = subprocess.Popen( cmd, shell=True, env=env,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE )
        output = p.communicate()[0]
        ok = p.returncode==0
    except OSError:
        ok = False
    sys.stdout.flush()
    sys.stderr.flush()
    return ok, output

def _which(cmd):
    """Function like BASH which (returns None if cmd not found)"""
    from distutils.spawn import find_executable
    cmd=find_executable(cmd)
    if not cmd:
        return None
    import os
    return os.path.abspath(os.path.realpath(cmd))

def _actual_test_cmdline( verbose ):
    prfct = _get_prfct( verbose )
    prfct('starting testing of cmd-line utilities')
    cmds = ['ncrystal-config --help',
            'ncrystal-config -s',
            'nctool --version',
            'nctool --help',
            'nctool --test',
            'ncrystal_inspectfile --help',
            'ncrystal_endf2ncmat --help',
            'ncrystal_hfg2ncmat --help',
            'ncrystal_ncmat2cpp --help',
            'ncrystal_ncmat2hkl --help',
            'ncrystal_cif2ncmat --help',
            'ncrystal_vdos2ncmat --help',
            'ncrystal_verifyatompos --help',
            'ncrystal-config --show cmakedir']
    for cmd in cmds:
        prfct('Trying to run:',cmd)
        ok, output = _run_cmd(cmd)
        if not ok:
            raise RuntimeError('Command failed: %s'%cmd)
    prfct('testing of cmd-line utilities done')

import contextlib as _contextlib
@_contextlib.contextmanager
def _work_in_tmpdir():
    """Context manager for working in a temporary directory (automatically created+cleaned) and then switching back"""
    import os
    import tempfile
    the_cwd = os.getcwd()
    try:
        with tempfile.TemporaryDirectory() as tmpdir:
            os.chdir(tmpdir)
            yield
    finally:
        os.chdir(the_cwd)

def _actual_test_cmake( verbose = False, ignore_if_absent = False ):
    prfct = _get_prfct( verbose )
    prfct('starting testing of compiled downstream cmake-based projects')
    cmakecmd = _which('cmake')
    if not cmakecmd:
        if ignore_if_absent:
            prfct('Skipping CMake test since "cmake" command not found.')
            return
        raise RuntimeError('cmake command not found')

    prfct('Using cmake command: %s'%cmakecmd)


    _singleline2str = lambda x : ((x.decode() if hasattr(x,'decode') else x) or '').strip()

    _cmd = 'ncrystal-config --show cmakedir'
    ok, ncrystal_cmakedir = _run_cmd(_cmd)
    if not ok:
        raise RuntimeError('Command failed: %s'%_cmd)
    ncrystal_cmakedir = _singleline2str(ncrystal_cmakedir)
    prfct('ncrystal-config gives cmakedir: %s'%ncrystal_cmakedir)
    _cmd = 'ncrystal-config --show libdir'
    ok, ncrystal_libdir = _run_cmd(_cmd)
    if not ok:
        raise RuntimeError('Command failed: %s'%_cmd)
    ncrystal_libdir = _singleline2str(ncrystal_libdir)
    prfct('ncrystal-config gives libdir: %s'%ncrystal_libdir)

    def prepend_to_path( envdict, pathvar, entry ):
        from shlex import quote
        entry = quote(((entry.decode() if hasattr(entry,'decode') else entry) or '').strip())
        if not entry:
            return
        l = [ entry ]
        orig = envdict.get(pathvar,'')
        if orig:
            l.append( orig )
        envdict[pathvar] = ':'.join(l)

    import os
    cmake_env = os.environ.copy()
    runtime_env = os.environ.copy()
    prfct('Testing with cmakedir added to CMAKE_PREFIX_PATH and libdir added to (DY)LD_LIBRARY_PATH')
    prepend_to_path( cmake_env, 'CMAKE_PREFIX_PATH', ncrystal_cmakedir )
    prepend_to_path( runtime_env, 'LD_LIBRARY_PATH', ncrystal_libdir )
    prepend_to_path( runtime_env, 'DYLD_LIBRARY_PATH', ncrystal_libdir )
    def runcmake(args):
        cmake_args = cmake_env.get('CMAKE_ARGS','')
        cmd=f'{cmakecmd} {cmake_args} {args}'
        ok, output = _run_cmd(cmd, env=cmake_env)
        prfct(f'Launching: {cmd}')
        if not ok:
            raise RuntimeError(f'Test command failed: {cmd}')

    downstream_cmake = """
cmake_minimum_required(VERSION 3.10...3.24)
project(MyExampleProject LANGUAGES CXX)
find_package(NCrystal REQUIRED)
file(WRITE "${PROJECT_BINARY_DIR}/cppsrc.cpp" "
#include \\"NCrystal/NCrystal.hh\\"
#include <iostream>
int main() {
  auto pc = NCrystal::createScatter( \\"Al_sg225.ncmat;dcutoff=0.5;temp=25C\\" );
  auto wl = NCrystal::NeutronWavelength{1.8};
  auto xsect = pc.crossSectionIsotropic( wl );
  std::cout << \\"Powder Al x-sect at \\" << wl << \\" is \\" << xsect << std::endl;
  return 0;
}
")
add_executable(exampleapp "${PROJECT_BINARY_DIR}/cppsrc.cpp")
target_link_libraries( exampleapp NCrystal::NCrystal )
install( TARGETS exampleapp DESTINATION bin )
"""

    import pathlib
    with _work_in_tmpdir():
        td = pathlib.Path('.').absolute()
        ( td / 'src' ).mkdir()
        ( td / 'src' / 'CMakeLists.txt' ).write_text(downstream_cmake)
        ( td  / 'bld' ).mkdir()
        os.chdir( td / 'bld' )
        runcmake('../src -DCMAKE_INSTALL_PREFIX=../install')
        runcmake('--build . --target install --config Release')
        os.chdir( td )
        app = ( td / 'install' / 'bin' / 'exampleapp' )
        if not app.exists():
            raise RuntimeError('Could not produce example app in downstream cmake project')
        prfct(f'Launching: {app}')
        ok, output = _run_cmd(str(app),env=runtime_env)
        if not ok:
            raise RuntimeError('Could not launch example app compiled in downstream cmake project')
        envrun = os.environ.copy()
        expected_output='Powder Al x-sect at 1.8Aa is 1.44816barn'
        if not output.decode().strip()==expected_output:
            raise RuntimeError('Example app compiled in downstream cmake project produces unexpected output')
        prfct(f'App produced expected output')
    prfct('testing of compiled downstream cmake-based projects done')




"""Internal module providing ctypes-based hooks into the compiled NCrystal shared library"""

################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

__all__ = ['_np',
           '_ensure_numpy',
           '_np_linspace',
           '_np_geomspace',
           '_np_logspace' ]

try:
    import numpy as _np
except ImportError:
    _np = None

def _ensure_numpy():
    if not _np:
        from .exceptions import NCException
        raise NCException("Numpy not available - array based functionality is unavailable")
    return _np

def _np_linspace(start,stop,num=50):
    """linspace with reproducible endpoint value"""
    _ensure_numpy()
    assert num >= 2
    l = _np.linspace(start,stop,num)
    l[0] = start
    l[-1] = stop
    return l

def _np_geomspace(start,stop,num=50):
    """geomspace with reproducible endpoint value"""
    _ensure_numpy()
    assert num >= 2
    l = _np.geomspace(start,stop,num)
    l[0] = start
    l[-1] = stop
    return l

def _np_logspace(start,stop,num=50):
    """logspace with reproducible endpoint value"""
    _ensure_numpy()
    assert num >= 2
    l = _np.logspace(start,stop,num)
    l[0] = 10.0**start
    l[-1] = 10.0**stop
    return l



"""

Module containing various McStas-related utilities, such as functions helping
with automatic setup of McStas-Union materials from NCrystal cfg-strings, or
production of .laz/.lau files for McStas components such as PowderN.comp or
Single_crystal.comp (with reduced physics capabilities of course).

"""

################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

from . import api as _NC
from ._common import warn as _nc_warn

def cfgstr_detect_components( cfgstr ):
    """
    Internal helper function which can detect which physics components are
    present once a given cfg-string is loaded, and provide derived cfg-strings
    suitable for picking out those components. The return value is a list of
    component names (e.g. "inelas", "cohelas", "sans", etc.) and the associated
    cfg-strings: [(cfgstr1,compname1), (cfgstr2,compname2),...].
    """
    #Normalise original cfg (so syntax errors will refer to the cfgstr as
    #originally specified), and add flags which cause initialisation speedup
    #without affecting which components are present:
    probecfgstr = _NC.normaliseCfg(cfgstr) + ';vdoslux=0'
    res=[]
    from . import misc as nc_misc
    for ct in nc_misc.standard_comp_types:
        extracfg = f';comp={ct}'
        if not _NC.createScatter( probecfgstr + extracfg ).isNull():
            res.append( ( _NC.normaliseCfg(cfgstr+extracfg), ct.replace('_','') ) )
    return res

def cfgstr_2_unioncfg( *, cfgstr, split_by_physics = False ):
    """Analyse cfg-string and return data needed in order to set up Union
    components accordingly. Specifically the function returns
    (inv_pen_depth_2200,proclist) where inv_pen_depth is the inverse penetration
    depth in inverse meters for a neutron at 2200m/s, and proclist is a list
    [(cfgstr1,compname1), (cfgstr2,compname2),...] with the relevant
    components. If split_by_physics is True, this process list will be as
    provided by cfgstr_detect_components(cfgstr), and otherwise it will simply
    be [(normaliseCfg(cfgstr),'total')].
    """
    info=_NC.createInfo(cfgstr)
    inv_pen_depth_2200 = 100.0 * info.getNumberDensity() * info.getXSectAbsorption()# in inverse meters
    if split_by_physics:
        return inv_pen_depth_2200, cfgstr_detect_components( cfgstr )
    else:
        return inv_pen_depth_2200, [(_NC.normaliseCfg(cfgstr),'total')]

def cfgstr_2_union_instrument_code( *, cfgstr, name, split_by_physics = False ):
    """
    Analyse cfg-string (via a call to cfgstr_2_unioncfg(cfgstr,split_by_physics)),
    and use the results to produce (and return) McStas-instrument code suitable
    for defining a McStas-Union material with the given name as specified via
    the "name" parameter. Depending on the value of split_by_physics, this Union
    material will either be set up with either just a single NCrystal_process,
    or a whole list of NCrystal_processes. Using split_by_physics=False is
    likely slightly more efficient, but split_by_physics=True will allow insight
    into the contributions of the various physics components ("inelas",
    "coh_elas", "sans", "incoh_elas", ...) at the McStas-Union level.
    """
    for ch in ';.:':
        if ch in name:
            raise ValueError(f'union process name "{name}" contains a "{ch}"'
                             +' character which is likely not what was intended.')
    assert not '/*' in cfgstr
    assert not '*/' in cfgstr
    assert not '/*' in name
    assert not '*/' in name

    out_absorption, physlist = cfgstr_2_unioncfg( cfgstr=cfgstr, split_by_physics=split_by_physics )
    res = f"""
/*
   The following code was auto generated by NCrystal v{_NC.get_version()} via Python:

     NCrystal.mcstasutils.cfgstr_2_union_instrument_code(
         cfgstr = {repr(str(cfgstr))},
         name = {repr(str(name))}"""
    if split_by_physics:
        res += """,
         split_by_physics = True )
"""
    else:
        res += ' )\n'
    res+="""
   Please rerun in case of major changes to input data or NCrystal.
*/
"""

    procnames=[]
    mcstas_string_max = 256
    for proccfgstr, proctypename in physlist:
        if len(physlist)==1 and proctypename=='total':
            procname=f'{name}_ncrystal_proc'
        else:
            procname=f'{name}_ncrystal_{proctypename}_proc'
        procnames.append( procname )
        if len(proccfgstr) > mcstas_string_max:
            _nc_warn(f'cfg_string might be too long for a McStas string: "{proccfgstr}"')

        res += f"""
COMPONENT {procname} = NCrystal_process(
    cfg = "{proccfgstr}" )
AT (0,0,0) ABSOLUTE
"""
    procnamestr = ','.join(procnames)
    if len(procnamestr) > mcstas_string_max:
        _nc_warn(f'process_string might be too long for a McStas string: "{procnamestr}" '
                 +'(it might work if you use a shorter material name)')
    res += f"""
COMPONENT {name} = Union_make_material(
    process_string = "{procnamestr}",
    my_absorption = {out_absorption:.15g} )
AT (0,0,0) ABSOLUTE

/* End of auto generated code from NCrystal v{_NC.get_version()}. */
"""



    return res

def cfgstr_2_hkl(*, cfgstr, tgtformat, verbose=True, fp_format = '%.14g', prefer_systemexit_exceptions = False ):
    """Function which can be used to create input files with reflections for
    McStas crystalline sample components like PowderN and Single_crystal, based
    on NCrystal cfg-strings (usually referring to NCMAT files with crystalline
    single-phase materials). The tgtformat must be either 'laz' or 'lau'. If
    verbose is True, the files might contain strictly unneccessary content
    (e.g. white-space for adjusting columns, a dspacing column in .lau files,
    ...). Finally the fp_format parameter can be used to change the precision of
    floating point numbers in the file.

    This function returns an iterable, yielding one line of the output file at a
    time.

    """
    import numbers
    import functools
    import math

    assert tgtformat in ('laz','lau')
    doPowder = (tgtformat=='laz')
    doDsp = (doPowder or verbose)
    doMult = True#Always needed, even in lau

    cfgstr = _NC.normaliseCfg(cfgstr)
    info = _NC.createInfo(cfgstr)
    def errmsg(msg):
        raise (SystemExit if prefer_systemexit_exceptions else RuntimeError)('Error: %s'%msg)

    if info.isMultiPhase():
        errmsg('this script does not handle multiphase materials')
    if not info.hasHKLInfo():
        errmsg('this script does not handle non-crystalline materials')
    if not info.hasStructureInfo():
        errmsg('this script does not handle crystalline materials without unit cell structure')
    si = info.getStructureInfo()
    if not si.get('spacegroup',None):
        errmsg('this script does not handle crystalline materials without spacegroup number')
    if not info.hasAtomInfo():
        errmsg('this script does not handle crystalline materials without info of atoms placement in the unit cell')
    if info.hklInfoType() != _NC.HKLInfoType.SymEqvGroup:
        errmsg('this script does not handle crystalline materials without symmetry equivalent HKL groupings')

    orig_header = []
    decoded_cfg = _NC.decodeCfg(cfgstr)
    data_name = decoded_cfg.get('data_name',None)
    cfgstr_nodataname = cfgstr.replace(data_name,'')
    if cfgstr_nodataname.startswith(';'):
        cfgstr_nodataname = cfgstr_nodataname[1:]
    cfgstr_nodataname = cfgstr_nodataname.strip()

    if ( decoded_cfg.get('density',{}).get('type',None) != 'scalefactor'
         or decoded_cfg.get('density',{}).get('value',None) != 1.0 ):
        errmsg('this script does not handle configurations with density overrides'
               +' (since resulting files might not load correctly by all clients)')

    if data_name:
        for i,l in enumerate(_NC.createTextData(data_name)):
            if l.startswith('#') or not l.strip():
                orig_header.append(l.rstrip())
            else:
                if not (i==0 and l.startswith('NCMAT')):
                    break
    assert info.hasStructureInfo()
    assert info.hasHKLInfo()
    assert info.hasTemperature()
    assert info.hasAtomInfo()

    si = info.getStructureInfo()
    fmtfp=lambda x : fp_format%x if isinstance(x, numbers.Real) else str(x)

    yield f'# File created by NCrystal v{_NC.get_version()}'
    yield '#'
    yield f'# ncrystal_cfgstr "{cfgstr}"'
    yield '#'
    if tgtformat=='laz':
        yield '# Format: "laz" (suitable for McStas PowderN component)'
    else:
        assert tgtformat=='lau'
        yield '# Format: "lau" (suitable for McStas Single_crystal component)'
    yield '#'

    if orig_header:
        had_embedded_cfg = any( 'NCRYSTALMATCFG[' in e for e in orig_header)
        guard=' orighdr :'#something that hopefully will cause McStas code to ignore the line
        yield f'#{guard} Original file had the following comments at the top:'
        if had_embedded_cfg:
            yield f'#{guard}'
            yield f'#{guard} (note the ncrystalmatcfg statement is disabled in this'
            yield f'#{guard}  reproduction by adding "<disable>" around it)'
        yield f'#{guard}'
        for l in orig_header:
            if 'NCRYSTALMATCFG[' in l:
                l=l.replace('NCRYSTALMATCFG[','NCRYSTAL<disable>MATCFG[')
            yield f'#{guard} {l}'
        yield '#'
    if cfgstr_nodataname:
        yield '#'
        yield f'# ncrystal_embedded_cfg : NCRYSTALMATCFG[{cfgstr_nodataname}]'
        yield '#'
    yield f'# temperature {fmtfp(info.getTemperature())} [kelvin]'
    yield '#'
    yield f'# spacegroup {si["spacegroup"]}'
    yield f'# lattice_a {fmtfp(si["a"])} [Aa]'
    yield f'# lattice_b {fmtfp(si["b"])} [Aa]'
    yield f'# lattice_c {fmtfp(si["c"])} [Aa]'
    yield f'# lattice_aa {fmtfp(si["alpha"])} [degrees]'
    yield f'# lattice_bb {fmtfp(si["beta"])} [degrees]'
    yield f'# lattice_cc {fmtfp(si["gamma"])} [degrees]'
    yield f'# Vc {fmtfp(si["volume"])} [Aa^3]'
    yield '#'
    n_atoms = sum( ai.count for ai in info.atominfos )
    assert si["n_atoms"] == n_atoms
    yield f'# multiplicity {n_atoms} [atoms/unit cell]'
    yield f'# density {fmtfp(info.density)} [g/cm^3]'
    yield f'# number_density {fmtfp(info.numberdensity)} [atoms/Aa^3]'
    daltons_per_unitcell = sum(ai.atomData.averageMassAMU() * ai.count for ai in info.atominfos)
    yield f'# weight {fmtfp(daltons_per_unitcell)} [g/mol of entire unit cell]'
    yield f'# average_mass {fmtfp(daltons_per_unitcell/n_atoms)} [average atomic g/mol]'

    sigma_coh = sum(ai.atomData.coherentXS() * ai.count for ai in info.atominfos)
    sigma_inc = sum(ai.atomData.incoherentXS() * ai.count for ai in info.atominfos)
    sigma_abs = sum(ai.atomData.captureXS() * ai.count for ai in info.atominfos)
    yield '#'
    yield f'# sigma_coh {fmtfp(sigma_coh)} [barn/unitcell]'
    yield f'# sigma_inc {fmtfp(sigma_inc)} [barn/unitcell]'
    yield f'# sigma_abs {fmtfp(sigma_abs)} [barn/unitcell]'
    yield '#'

    has_debye_temp = True
    debye_temp_sum = 0.0
    debye_temp_sumw = 0.0

    if all(ai.atomData.isElement() for ai in info.atominfos):
        d = {}
        for ai in info.atominfos:
            if has_debye_temp:
                if ai.debyeTemperature is not None:
                    _dt = ai.debyeTemperature
                elif ai.msd is not None:
                    _dt = _NC.debyeTempFromIsotropicMSD( msd = ai.msd,
                                                        temperature = info.getTemperature(),
                                                        mass = ai.atomData.averageMassAMU() )
                else:
                    has_debye_temp = False
                    _dt = 0.0
                _dtw = ai.count * ai.atomData.scatteringXS()
                debye_temp_sum += _dt*_dtw
                debye_temp_sumw += _dtw

            d[ai.atomData.elementName()] = d.get(ai.atomData.elementName(),0) + ai.count
            nformula_per_unitcell = functools.reduce(math.gcd, list(c for _,c in d.items()))
        formula = ''.join(( '%s%i'%(k,v/nformula_per_unitcell) if v!=nformula_per_unitcell else k) for k,v in sorted(d.items()))
        yield f'# formula {formula}'
        yield f'# nformula_per_unitcell {nformula_per_unitcell}'
        if debye_temp_sum:
            yield f'# debye_temperature {fmtfp(debye_temp_sum/debye_temp_sumw)} [kelvin, weighted by sigma_b]'

    cols = ['h','k','l']
    if doMult:
        cols.append('j')
    if doDsp:
        cols.append('d')
    cols.append('F2')

    colwidths={'h':3,'k':3,'l':3,'j':2,'d':16,'F2':16}
    header = '#'
    for i,c in enumerate(cols):
        header += ' '
        if verbose:
            header += c.rjust(colwidths[c])
        else:
            header += c

    col_comments={
        'd':"[d-spacing in Aa]",
        'j': "multiplicity",
        'F2':"[norm of scattering factor |F|^2 in barn]"
    }

    yield '#'
    maxcw=max(len(c) for c in col_comments.keys())
    for i,c in enumerate(cols):
        cc = col_comments.get(c,'')
        if cc:
            cc = '  '+cc
        yield f'# column_{c.ljust(maxcw)} {i+1}{cc}'
    yield '# unit_F2 barn'
    yield '#'

    yield header
    def format_line(data):
        s=' ' if verbose else ''
        for c in cols:
            if s:
                s+=' '
            s+=fmtfp(data[c]).rjust(colwidths[c])
        return s

    if doPowder:
        for h,k,l,mult,dsp,F2_barn in info.hklList():
            yield format_line({'h':h,'k':k,'l':l,'d':dsp,'j':mult,'F2':F2_barn})
    else:
        for h,k,l,mult,dsp,F2_barn in info.hklList(all_indices=True):
            for i in range(len(h)):
                #Yield both hkl and -hkl and put mult=1 (since we put all planes explicitly)
                yield format_line({'h':h[i],'k':k[i],'l':l[i],'d':dsp,'j':1,'F2':F2_barn})
                yield format_line({'h':-h[i],'k':-k[i],'l':-l[i],'d':dsp,'j':1,'F2':F2_barn})

def main(argv):
    args = argv[1:]
    if args and isinstance(args[0],bytes):
        args = list(e.decode() for e in args)
    def usage(*,err):
        if err:
            print("ERROR - wrong usage!")
            print()
        import os
        pn=os.path.basename(argv[0])
        if pn.endswith('.py'):
            pn = pn[0:-3]
        print("""Usage:

  --help|-h                         Show these instructions

  --union [--split] NAME CFGSTR     Output McStas-Union code to define material
                                    with NAME based on specified NCrystal CFGSTR.
                                    Providing --split will split NCrystal processes
                                    into physics types at the McStas-Union level,
                                    rather than internally in NCrystal.

  --laz CFGSTR                      Convert NCrystal CFGSTR to McStas .laz format
                                    for PowderN.

  --lau CFGSTR                      Convert NCrystal CFGSTR to McStas .lau format
                                    for Single_crystal.

""")
        raise SystemExit(1 if err else 0)
    err = lambda : usage(err=True)
    if '-h' in args or '--help' in args:
        usage(err=False)
    if not args or len(args)<2:
        err()
    if args[0] == '--union':
        dosplit = False
        if '--split' in args:
            args.remove('--split')
            dosplit=True
        if len(args)<3:
            err()
        print(cfgstr_2_union_instrument_code( cfgstr=';'.join(args[2:]), name=args[1], split_by_physics = dosplit ))
        return
    if args[0] in ('--laz','--lau'):
        for e in cfgstr_2_hkl( cfgstr=';'.join(args[1:]),tgtformat=args[0][2:], prefer_systemexit_exceptions = True):
            print(e)
        return
    err()

if __name__ == '__main__':
    import sys
    main(sys.argv)




################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

"""

Utilities for accessing NCrystal's database of atomic data, with information
about atomic masses, scattering lengths, etc. Also contains a few other related
utilities, like a list of all element names.

"""

def normaliseCfg(cfgstr):
    """Returns normalised version of cfgstr. This is done behind the scenes by
       loading the specified cfg-string into a C++ MatCfg object and then
       re-encoding it as a string.
    """
    from ._chooks import _get_raw_cfcts
    return _get_raw_cfcts()['nc_normcfgstr'](cfgstr)

def decodeCfg(cfgstr,*,asJSONStr=False):
    """Decodes cfg-string and returns as Python data structure (a dictionary). The
       format of this data structure should be mostly self-evident by
       inspection, and is not guaranteed to stay the same across NCrystal
       versions. If asJSONStr=true, the data structure will be returned as a
       JSON-encoded string, instead of a Python dictionary.
    """
    from ._chooks import _get_raw_cfcts
    _js = _get_raw_cfcts()['nc_cfgstr2json'](cfgstr)
    if asJSONStr:
        return _js
    import json
    return json.loads(_js)

def generateCfgStrDoc( mode = "print" ):
    """Generates documentation about the available cfg-str variables. Mode can
    either be 'print' (print detailed explanation to stdout), 'txt_full' (return
    detailed explanations as string), 'txt_short' (return short explanations as
    string), 'json' (return json-encoded string), or 'python' (return python
    data structures).
    """
    modemap={'print':0,'txt_full':0,'txt_short':1,'json':2,'python':2}
    modeint = modemap.get(mode,None)
    if modeint is None:
        from .exceptions import NCBadInput
        raise NCBadInput('mode must be one of %s'%list(sorted(modemap.keys())))
    from ._chooks import _get_raw_cfcts
    _=_get_raw_cfcts()['nc_gencfgdoc'](modeint)
    if mode == 'print':
        print(_)
    elif mode == 'python':
        import json
        return json.loads(_)
    else:
        return _

def decodecfg_vdoslux(cfgstr):
    """Extract vdoslux value from cfgstr."""
    from ._chooks import _get_raw_cfcts,_str2cstr
    return int(_get_raw_cfcts()['ncrystal_decodecfg_vdoslux'](_str2cstr(cfgstr)))



#!/usr/bin/env python3
"""Python module for using the NCrystal library for thermal neutron transport in crystals and other materials.

Please find more information about NCrystal at the website:

   https://mctools.github.io/ncrystal/

In particular, a small example using the NCrystal python module can be found at:

   https://github.com/mctools/ncrystal/blob/master/examples/ncrystal_example_py

A substantial effort went into developing NCrystal. If you use it for your work,
we would appreciate it if you would use the following reference in your work:

  X.-X. Cai and T. Kittelmann, NCrystal: A library for thermal neutron
  transport, Computer Physics Communications 246 (2020) 106851,
  https://doi.org/10.1016/j.cpc.2019.07.015

For work benefitting from our inelastic physics, we furthermore request that you
additionally also use the following reference in your work:

  X.-X. Cai, T. Kittelmann, et. al., "Rejection-based sampling of inelastic
  neutron scattering", Journal of Computational Physics 380 (2019) 400-407,
  https://doi.org/10.1016/j.jcp.2018.11.043

For detailed usage conditions and licensing of this open source project, see:

   https://github.com/mctools/ncrystal/blob/master/NOTICE
   https://github.com/mctools/ncrystal/blob/master/LICENSE

"""

################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

#NB: Synchronize meta-data below with fields in setup.py+template_setup.py.in meta data:
__license__ = "Apache 2.0, http://www.apache.org/licenses/LICENSE-2.0"
__version__ = '3.7.1'
__status__ = "Production"
__author__ = "NCrystal developers (Thomas Kittelmann, Xiao Xiao Cai)"
__copyright__ = "Copyright 2015-2023 %s"%__author__
__maintainer__ = __author__
__email__ = "ncrystal-developers@cern.ch"

#Place f-string here to catch python <3.6 in a more obvious way than a syntax error below:
f'NCrystal does not work with Python2 (or Python3 < v3.6)'
_minpyversion=(3,6,0)

import sys as _sys
pyversion = _sys.version_info[0:3]
if pyversion < _minpyversion:
    raise SystemExit('Unsupported python version %i.%i.%i detected (needs %i.%i.%i or later).'%(pyversion+_minpyversion))

import os as _os
if not _os.environ.get('NCRYSTAL_SLIMPYINIT'):
    from .api import *

###################################
#Same as NCRYSTAL_VERSION macro (same as the get_version_num() function from .core):
version_tuple = tuple( int(i) for i in __version__.split('.') )#introduced in release 3.6.0
version_num = sum(int(i)*j for i,j in zip(version_tuple,(1000000,1000,1)))#Introduced in release 0.9.1
